from abc import ABC, abstractclassmethod, abstractmethod
from typing import Any, Optional, Sequence, Tuple, Union

import numpy as np
import pandas as pd

from ..check import Check, ColumnCheck


class _Metric(ABC):
    name: Optional[str] = None
    tags: Sequence[str] = []

    def __repr__(self):
        return f"{self.__class__.__name__}()"

    def __str__(self):
        return f"{self.name}"


class OneColumnMetric(_Metric):

    def __init__(self, check: Check = ColumnCheck()):
        self.check = check

    @abstractclassmethod
    def check_column_types(cls, sr: pd.Series, check: Check = ColumnCheck()) -> bool:
        ...

    @abstractmethod
    def _compute_metric(self, sr: pd.Series):
        ...

    def __call__(self, sr: pd.Series):
        if not self.check_column_types(sr, self.check):
            return None
        return self._compute_metric(sr)


class TwoColumnMetric(_Metric):
    def __init__(self, check: Check = ColumnCheck()):
        self.check = check

    @abstractclassmethod
    def check_column_types(cls, sr_a: pd.Series, sr_b: pd.Series, check: Check = ColumnCheck()):
        ...

    @abstractmethod
    def _compute_metric(self, sr_a: pd.Series, sr_b: pd.Series):
        ...

    def __call__(self, sr_a: pd.Series, sr_b: pd.Series):
        if not self.check_column_types(sr_a, sr_b, self.check):
            return None
        return self._compute_metric(sr_a, sr_b)


class TwoColumnTest(_Metric):
    def __init__(self,
                 check: Check = ColumnCheck(),
                 alternative: str = 'two-sided'):

        if alternative not in ('two-sided', 'greater', 'less'):
            raise ValueError("'alternative' argument must be one of 'two-sided', 'greater', 'less'")

        self.check = check
        self.alternative = alternative

    @abstractclassmethod
    def check_column_types(cls, sr_a: pd.Series, sr_b: pd.Series, check: Check = ColumnCheck()) -> bool:
        ...

    @abstractmethod
    def _compute_test(self,
                      sr_a: pd.Series,
                      sr_b: pd.Series) -> Tuple[Union[int, float, None], Union[int, float, None]]:
        ...

    def __call__(self, sr_a: pd.Series, sr_b: pd.Series) -> Tuple[Union[int, float, None], Union[int, float, None]]:
        if not self.check_column_types(sr_a, sr_b, self.check):
            return None, None

        return self._compute_test(sr_a, sr_b)


class ModellingMetric(_Metric):

    @abstractmethod
    def __call__(self,
                 y_true: np.ndarray,
                 y_pred: Optional[np.ndarray] = None) -> Union[float, None]:
        pass


class ClassificationMetric(ModellingMetric):
    tags = ["modelling", "classification"]
    plot = False

    def __init__(self, multiclass: bool = False):
        self.multiclass = multiclass

    def __call__(self, y_true: np.ndarray, y_pred: Optional[np.ndarray] = None,
                 y_pred_proba: Optional[np.ndarray] = None) -> float:
        raise NotImplementedError


class ClassificationPlotMetric(ModellingMetric):
    tags = ["modelling", "classification", "plot"]
    plot = True

    def __init__(self, multiclass: bool = False):
        self.multiclass = multiclass

    def __call__(self, y_true: np.ndarray, y_pred: Optional[np.ndarray] = None,
                 y_pred_proba: Optional[np.ndarray] = None) -> Any:
        raise NotImplementedError


class RegressionMetric(ModellingMetric):
    tags = ["modelling", "regression"]

    def __init__(self):
        # Contains nothing atm but matches other two Modelling metrics.
        pass

    def __call__(self, y_true: np.ndarray, y_pred: np.ndarray = None) -> float:
        raise NotImplementedError


class DataFrameMetric(_Metric):

    @abstractmethod
    def __call__(self, df: pd.DataFrame) -> Union[int, float, None]:
        pass


class TwoDataFrameMetric(_Metric):

    @abstractmethod
    def __call__(self, df_old: pd.DataFrame, df_new: pd.DataFrame) -> Union[int, float, None]:
        pass



from typing import NamedTuple, Tuple, Union, cast

import numpy as np
import pandas as pd
from scipy.stats import beta, norm

from .base import OneColumnMetric, TwoColumnMetric
from .utils import bootstrap_binned_statistic, bootstrap_statistic


class ConfidenceInterval(NamedTuple):
    limits: Tuple
    level: float


def compute_bootstrap_interval(metric_cls_obj: Union[OneColumnMetric, TwoColumnMetric],
                               sr_a: pd.Series,
                               sr_b: pd.Series = None,
                               confidence_level: float = 0.95,
                               binned: bool = False) -> ConfidenceInterval:
    """Return a frequentist confidence interval for this metric obtained, via bootstrap resampling

    Args:
        sr_a: Value of one column
        sr_b: Value of another column
        metric_cls_obj: Instantiated object of a metric class
        confidence_level: Level on which confidence interval is computed
        binned: True in case of binning

    Returns:
        The confidence interval.
    """

    if isinstance(metric_cls_obj, OneColumnMetric):
        metric_value = metric_cls_obj(sr_a)
        one_col_metric: OneColumnMetric = metric_cls_obj  # Need explicit casting because of mypy bug/issue (#2608)
        samples = bootstrap_statistic((sr_a,), lambda x: one_col_metric(x))
    else:
        metric_value = metric_cls_obj(sr_a, cast(pd.Series, sr_b))
        two_col_metric: TwoColumnMetric = metric_cls_obj  # Need explicit casting because of mypy bug/issue (#2608)
        if not binned:
            samples = bootstrap_statistic((sr_a, cast(pd.Series, sr_b)), lambda x, y: two_col_metric(x, y))
        else:
            samples = bootstrap_binned_statistic((sr_a, cast(pd.Series, sr_b)), lambda x, y: two_col_metric(x, y))

    percentiles = 100 * (1 - confidence_level) / 2, 100 * (1 - (1 - confidence_level) / 2)
    d1, d2 = np.percentile(samples, percentiles)
    return ConfidenceInterval(limits=(2 * metric_value - d2, 2 * metric_value - d1), level=confidence_level)


def binominal_proportion_interval(success_prop: float,
                                  num_samples: int,
                                  method: str = 'clopper-pearson',
                                  confidence_level: float = 0.95) -> ConfidenceInterval:
    """
    Calculate an approximate confidence interval for a binomial proportion of a sample.
    Should only be used for binomial distribution

    Args:
        success_prop: Proportion of sucesses.
        num_samples: Sample size.
        method: Optional; The approximation method used to calculate the interval.
            One of 'normal', 'clopper-pearson', 'agresti-coull'.
        confidence_level: Level on which confidence interval is computed

    Returns:
        A ConfidenceInterval containing the interval and confidence level.
    """

    k = num_samples * success_prop
    alpha = 1 - confidence_level
    z = norm.ppf(1 - alpha / 2)

    if method == 'normal':
        low = success_prop - z * np.sqrt(success_prop * (1 - success_prop) / num_samples)
        high = success_prop + z * np.sqrt(success_prop * (1 - success_prop) / num_samples)

    elif method == 'clopper-pearson':
        low = beta.ppf(alpha / 2, k, num_samples - k + 1)
        high = beta.ppf(1 - alpha / 2, k + 1, num_samples - k)

    elif method == 'agresti-coull':
        n_ = num_samples + z**2
        p_ = 1 / n_ * (k + z**2 / 2)
        low = p_ - z * np.sqrt(p_ * (1 - p_) / n_)
        high = p_ + z * np.sqrt(p_ * (1 - p_) / n_)

    else:
        raise ValueError("'method' argument must be one of 'normal', 'clopper-pearson', 'agresti-coull'.")

    return ConfidenceInterval(limits=(low, high), level=confidence_level)


def compute_binomial_interval(sr_a: pd.Series,
                              sr_b: pd.Series,
                              method: str = 'clopper-pearson',
                              confidence_level: float = 0.95) -> ConfidenceInterval:
    """
    Calculate a confidence interval for this distance metric.
    Args:
        sr_a: value of a binary variable
        sr_b: value of a binary variable
        method: Optional; default is 'clopper-pearson'
        confidence_level: Level on which confidence interval is computed

    Returns:
        The confidence interval.
    """
    p = sr_a.mean()
    n = len(sr_a)
    interval = binominal_proportion_interval(p, n, method, confidence_level)
    cinterval = ConfidenceInterval((interval.limits[0] - sr_b.mean(), interval.limits[1] - sr_b.mean()), interval.level)
    return cinterval




from itertools import permutations
from typing import Optional, Tuple, Union

import numpy as np
import pandas as pd

from .base import TwoColumnMetric, TwoColumnTest


class TwoColumnMap:
    """Compares columns with the same name from two given dataframes and return a DataFrame
    with index as the column name and the columns as metric_val and metric_pval(if applicable)"""

    def __init__(self, metric: Union[TwoColumnMetric, TwoColumnTest]):
        self.metric = metric
        self.name = f'{metric.name}_map'

    def __call__(self, df_old: pd.DataFrame, df_new: pd.DataFrame) -> pd.DataFrame:
        columns_map = {col: self.metric(df_old[col], df_new[col]) for col in df_old.columns}

        result = pd.DataFrame(
            data=columns_map.values(),
            index=df_old.columns,
            columns=['metric_val', 'metric_pval'] if isinstance(self.metric, TwoColumnTest) else ['metric_val']
        )

        result.name = self.metric.name
        return result


class CorrMatrix:
    """Computes the correlation between each pair of columns in the given dataframe
    and returns the result in a dataframe"""

    def __init__(self, metric: Union[TwoColumnMetric, TwoColumnTest]):
        self.metric = metric
        self.name = f'{metric.name}_matrix'

    def __call__(self, df: pd.DataFrame) -> Tuple[pd.DataFrame, Optional[pd.DataFrame]]:
        columns = df.columns
        matrix = pd.DataFrame(index=columns, columns=columns)

        for col_a, col_b in permutations(columns, 2):
            matrix[col_a][col_b] = self.metric(df[col_a], df[col_b])

        pval_matrix = None
        if isinstance(self.metric, TwoColumnTest):
            value_matrix = pd.DataFrame(index=columns, columns=columns)
            pval_matrix = pd.DataFrame(index=columns, columns=columns)

            for col_a, col_b in permutations(columns, 2):
                value_matrix[col_a][col_b], pval_matrix[col_a][col_b] = matrix[col_a][col_b]
        else:
            value_matrix = matrix

        return pd.DataFrame(value_matrix.astype(np.float32)), pval_matrix  # explicit casting for mypy


class DiffCorrMatrix:
    """Computes the correlation matrix for each of the given dataframes and return the difference
    between these matrices"""

    def __init__(self, metric: Union[TwoColumnMetric, TwoColumnTest]):
        self.corr_matrix = CorrMatrix(metric)
        self.name = f'diff_{metric.name}'

    def __call__(self, df_old: pd.DataFrame, df_new: pd.DataFrame) -> Union[pd.DataFrame, None]:
        corr_matrix_old = self.corr_matrix(df=df_old)[0]
        corr_matrix_new = self.corr_matrix(df=df_new)[0]

        if corr_matrix_old is None or corr_matrix_new is None:
            return None

        return corr_matrix_old - corr_matrix_new



from .preprocessor import ModellingPreprocessor
from .utility import CLASSIFIERS, REGRESSORS, check_model_type, sample_split_data

__all__ = ['CLASSIFIERS', 'REGRESSORS', 'ModellingPreprocessor',
           'check_model_type', 'sample_split_data']



from abc import ABC, abstractmethod
from typing import cast

import numpy as np
import pandas as pd


class Check(ABC):
    """Class for checking column types
    Implement the methods of this class for the customized logic for different types of checks
    """

    def infer_dtype(self, sr: pd.Series) -> pd.Series:
        """Infers the type of the data and converts the data to it.
        Args:
            sr (str):
                The column of the dataframe to transform.
        Returns:
            pd.Series:
                The column converted to its inferred type.
        """

        col = sr.copy()

        in_dtype = str(col.dtype)
        n_nans = col.isna().sum()
        n_empty_str = 0 if not pd.api.types.is_string_dtype(col) else sr.eq("").sum()

        # Try to convert it to numeric
        if col.dtype.kind not in ("i", "u", "f") and col.dtype.kind != 'M':
            col_num = pd.to_numeric(col, errors="coerce")
            if col_num.isna().sum() == n_nans + n_empty_str:
                col = col_num

        # Try to convert it to date
        if col.dtype.kind == "O" or col.dtype.kind == 'M':
            try:
                col_date = pd.to_datetime(col, errors="coerce")
            except TypeError:
                col_date = pd.to_datetime(col.astype(str), errors="coerce")
            if col_date.isna().sum() == n_nans + n_empty_str:
                col = col_date

        out_dtype = str(col.dtype)

        if out_dtype == in_dtype:
            return col
        elif out_dtype in ("i", "u", "f", "f8", "i8", "u8"):
            return pd.to_numeric(col, errors="coerce")

        return cast(pd.Series, col.astype(out_dtype, errors="ignore"))

    @abstractmethod
    def continuous(self, sr: pd.Series):
        """Checks if the given series is continuous"""
        pass

    @abstractmethod
    def categorical(self, sr: pd.Series):
        """Checks if the given series is categorical"""
        pass

    @abstractmethod
    def ordinal(self, sr: pd.Series):
        """Checks if the given series is ordinal
        A series whose values can be ordered is ordinal
        """
        pass

    @abstractmethod
    def affine(self, sr: pd.Series):
        """Checks if the given series is affine"""
        pass


class ColumnCheck(Check):
    """Utility to check the column or a pairs of columns for different conditions
    Args:
        min_num_unique (int, optional):
            Minimum number of unique values for the data to be continuous.
            Default value is 10.
        ctl_mult (float, optional):
            Categorical threshold log multiplier.
            Default value is 2.5.
    """
    def __init__(self,
                 min_num_unique: int = 10,
                 ctl_mult: float = 2.5):
        self.min_num_unique = min_num_unique
        self.ctl_mult = ctl_mult

    def continuous(self, sr: pd.Series) -> bool:
        """Checks if the given series is continuous
        All arithmetic operations can be done on continuous columns
        """
        sr = self.infer_dtype(sr)
        sr_dtype = str(sr.dtype)
        if len(sr.unique()) >= max(self.min_num_unique,
                                   self.ctl_mult * np.log(len(sr)))\
           and sr_dtype in ("float64", "int64"):
            return True
        return False

    def categorical(self, sr: pd.Series) -> bool:
        """Checks if the given series is categorical"""
        if pd.api.types.is_categorical_dtype(sr) is True:
            return True
        sr = self.infer_dtype(sr)
        if sr.dtype.kind == "M":  # TODO: Need to implement ability to deal with dates well in metrics
            return False

        if not self.continuous(sr):
            return True
        return False

    def ordinal(self, sr: pd.Series) -> bool:
        """Checks if the given series is ordinal
        A series whose values can be ordered is ordinal
        Columns which are categorical in nature, but contain numbers/dates/bool
        are ordinal too. E.g. [2, 1, 1, 2, 7, 2, 1, 7]
        """
        if (pd.api.types.is_categorical_dtype(sr) is True
            and sr.cat.ordered is True)\
           or pd.api.types.is_bool_dtype(sr) is True:
            return True

        sr_inferred = self.infer_dtype(sr)
        if sr_inferred.dtype in ("float64", "int64")\
           or sr_inferred.dtype.kind in ("M", "m"):
            return True
        return False

    def affine(self, sr: pd.Series) -> bool:
        """Checks if the given series is affine
        Continuous columns along with the columns of type DateTime
        and the Timedelta are affine
        """
        if self.continuous(sr) is True\
           or self.infer_dtype(sr).dtype.kind in ("M", "m"):
            return True
        return False



import logging
from typing import Dict, List, Optional, Sequence, Union, cast

import numpy as np
import pandas as pd
from sklearn.base import BaseEstimator
from sklearn.impute import SimpleImputer
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import LabelEncoder, OneHotEncoder, StandardScaler

from ..check import Check, ColumnCheck

logger = logging.getLogger(__name__)


class ModellingPreprocessor:
    def __init__(self, target: Optional[str], check: Check = ColumnCheck()):
        self.target = target
        self.column_encoders: Dict[str, BaseEstimator] = dict()
        self.columns_mapping: Dict[str, List[str]] = dict()
        self.is_fitted: bool = False
        self.check = check
        self.continuous_cols: Union[None, Sequence[str]] = None
        self.categorical_cols: Union[None, Sequence[str]] = None

    def _continuous(self, df) -> Sequence[str]:
        if self.continuous_cols is None:
            self.continuous_cols = [col for col in df.columns if self.check.continuous(df[col])]
        return self.continuous_cols

    def _categorical(self, df) -> Sequence[str]:
        if self.categorical_cols is None:
            self.categorical_cols = [col for col in df.columns if self.check.categorical(df[col])]
        return self.categorical_cols

    def fit(self, df: pd.DataFrame):
        if self.is_fitted:
            logger.info("Preprocessor has already been fitted.")
            return

        for col in df.columns:
            df[col] = self.check.infer_dtype(df[col])

        for cat_col in self._categorical(df):
            logger.debug(f"Preprocessing value '{cat_col}'...")
            column = df[cat_col]
            column = cast(pd.Series, column.fillna('nan').astype(str))
            if self.target and cat_col == self.target:
                x_i = column.to_numpy().reshape(-1, 1)
                encoder = LabelEncoder()
                encoder.fit(x_i)
                c_name_i = [cat_col]

            else:
                x_i = column.to_numpy().reshape(-1, 1)
                encoder = OneHotEncoder(drop='first', sparse=False)
                encoder.fit(x_i)
                c_name_i = cast(List[str],
                                ['{}_{}'.format(cat_col, enc)
                                 for enc in encoder.categories_[0][1:]])

            self.column_encoders[cat_col] = encoder
            self.columns_mapping[cat_col] = c_name_i

        for cont_col in self._continuous(df):
            logger.debug(f"Preprocessing value '{cont_col}'...")
            column = df[cont_col]
            x_i = pd.to_numeric(column, errors="coerce").to_numpy()\
                                                        .reshape(-1, 1)

            n_rows = len(df)
            nan_freq = df[cont_col].isna().sum() / n_rows if n_rows > 0 else 0
            if nan_freq is not None and nan_freq > 0:
                self.column_encoders[cont_col] = Pipeline([('imputer', SimpleImputer()),
                                                           ('scaler', StandardScaler())])
            else:
                self.column_encoders[cont_col] = StandardScaler()

            self.column_encoders[cont_col].fit(x_i)
            self.columns_mapping[cont_col] = [cont_col]

        self.is_fitted = True

    def transform(self, df: pd.DataFrame) -> pd.DataFrame:
        if not self.is_fitted:
            raise ValueError("Preprocessor has not been fit yet.\
                              Call 'fit()' before calling 'transform()'")
        df = df.copy()

        xx = []
        c_names: List[str] = []

        for cat_col in self._categorical(df):
            column = df[cat_col]
            column = cast(pd.Series, column.fillna('nan').astype(str))
            encoder = self.column_encoders[cat_col]
            if self.target and cat_col == self.target:
                x_i = encoder.transform(column.values.reshape(-1, 1)).reshape(-1, 1)
                c_name_i = [cat_col]
            else:
                x_i = encoder.transform(column.values.reshape(-1, 1))
                c_name_i = ['{}_{}'.format(cat_col, enc)
                            for enc in encoder.categories_[0][1:]]

            xx.append(x_i)
            c_names.extend(c_name_i)

        for cont_col in self._continuous(df):
            if cont_col not in df.columns:
                continue
            column = df[cont_col]
            x_i = pd.to_numeric(column, errors="coerce").to_numpy()\
                                                        .reshape(-1, 1)
            x_i = self.column_encoders[cont_col].transform(x_i)
            c_name_i = [cont_col]
            xx.append(x_i)
            c_names.extend(c_name_i)

        if len(xx) == 0 or len(c_names) == 0:
            return pd.DataFrame()

        return pd.DataFrame(np.hstack(xx), columns=c_names)

    def fit_transform(self, df: pd.DataFrame) -> pd.DataFrame:
        self.fit(df)
        return self.transform(df)

    @property
    def processed_columns(self):
        if not self.is_fitted:
            raise ValueError("Preprocessor has not been fit yet.\
                        Call 'fit()' before accessing 'processed_columns'")

        return np.concatenate([self.columns_mapping.values()])

    @classmethod
    def preprocess(cls,
                   df: pd.DataFrame,
                   target: Optional[str]) -> pd.DataFrame:
        preprocessor = cls(target=target)
        return preprocessor.fit_transform(df)



import numpy as np
import pandas as pd
from sklearn.linear_model import LogisticRegression, Ridge, RidgeClassifier

from synthesized_insight.metrics import (
    Accuracy,
    ConfusionMatrix,
    F1Score,
    MeanAbsoluteError,
    MeanSquaredError,
    PRCurve,
    Precision,
    PredictiveModellingScore,
    R2Score,
    Recall,
    ROCCurve,
)
from synthesized_insight.metrics.modelling_metrics import (
    classifier_scores,
    classifier_scores_from_df,
    predictive_modelling_score,
    regressor_scores_from_df,
)

np.random.seed(42)


def test_predictive_modelling():
    x = np.random.normal(0, 1, 100)
    y = 2 * x + np.random.normal(0, 0.1, 100)
    y2 = 2.2 * x + np.random.normal(0, 0.1, 100)
    df_old = pd.DataFrame({'x': x, 'y': y})
    df_new = pd.DataFrame({'x': x, 'y': y2})

    pms = PredictiveModellingScore(model='Linear', y_label='y', x_labels=['x'], df_test=df_new)
    score, synth_score, _, _ = pms(df_old)
    assert score is not None
    assert synth_score is not None


def test_predictive_modelling_score_clf():
    n = 25
    data = pd.DataFrame({
        'x1': np.random.randn(n),
        'x2': np.where(np.random.uniform(size=n) < 0.8,
                       np.random.randn(n), np.nan),
        'x3': np.random.choice(['a', 'b', 'c', 'd'], size=n),
        'x4': np.where(np.random.uniform(size=n) < 0.8,
                       np.random.randint(1e5, size=n).astype(str), ''),
        'y': np.random.choice([0, 1], size=n)
    })

    target = 'y'
    x_labels = list(filter(lambda c: c != target, data.columns))
    score, metric, task = predictive_modelling_score(data,
                                                     model='Logistic',
                                                     y_label=target,
                                                     x_labels=x_labels)
    assert metric == 'roc_auc'
    assert task == 'binary'


def test_predictive_modelling_score_rgr():
    n = 1000
    data = pd.DataFrame({
        'x1': np.random.randn(n),
        'x2': np.where(np.random.uniform(size=n) < 0.8,
                       np.random.randn(n), np.nan),
        'x3': np.random.choice(['a', 'b', 'c', 'd'], size=n),
        'y': np.random.randn(n)
    })

    target = 'y'
    score, metric, task = predictive_modelling_score(data,
                                                     model='Linear',
                                                     y_label=target,
                                                     x_labels=None)
    assert metric == 'r2_score'
    assert task == 'regression'


# y_true has class 3, but y_pred doesn't
def test_classifier_scores_multiclass_classnotpresent_ypred():
    train = pd.DataFrame({
        'x1': ['100', '100', '100', '200', '200', '300'],
        'x2': [1, 2, 3, 1, 4, 5],
        'x3': [4, 8, 12, 3, 12, 9],
        'y': [1, 1, 1, 2, 2, 3]
    })

    test = pd.DataFrame({
        'x1': ['100', '200', '200'],
        'x2': [4, 2, 5],
        'x3': [16, 6, 15],
        'y': [1, 2, 3]
    })

    x_train = train[['x1', 'x2', 'x3']]
    y_train = train['y']
    x_test = test[['x1', 'x2', 'x3']]
    y_test = test['y']

    clf = RidgeClassifier()
    clf.fit(x_train, y_train)
    y_pred = clf.predict(x_test)
    assert (y_pred == np.array([1, 2, 2])).all()
    assert (np.array(y_test) == np.array([1, 2, 3])).all()

    metrics_dict = classifier_scores(x_train=np.array(x_train),
                                     y_train=np.array(y_train),
                                     x_test=np.array(x_test),
                                     y_test=np.array(y_test), clf=clf)

    assert np.isclose(metrics_dict['roc_auc'], 0.75, rtol=0.05)


# y_true and y_pred both don't have class 4
def test_classifier_scores_multiclass_classnotpresent_both():
    train = pd.DataFrame({
        'x1': ['100', '100', '100', '200', '200', '300', '500'],
        'x2': [1, 2, 3, 1, 4, 5, 11],
        'x3': [4, 8, 12, 3, 12, 9, 4],
        'y': [1, 1, 1, 2, 2, 3, 4]
    })

    test = pd.DataFrame({
        'x1': ['100', '200', '200'],
        'x2': [4, 2, 5],
        'x3': [16, 6, 15],
        'y': [2, 3, 1]
    })

    x_train = train[['x1', 'x2', 'x3']]
    y_train = train['y']
    x_test = test[['x1', 'x2', 'x3']]
    y_test = test['y']

    clf = RidgeClassifier()
    clf.fit(x_train, y_train)
    y_pred = clf.predict(x_test)
    assert (y_pred == np.array([1, 2, 3])).all()
    assert (np.array(y_test) == np.array([2, 3, 1])).all()

    metrics_dict = classifier_scores(x_train=np.array(x_train),
                                     y_train=np.array(y_train),
                                     x_test=np.array(x_test),
                                     y_test=np.array(y_test), clf=clf)

    assert np.isclose(metrics_dict['roc_auc'], 0.25, rtol=0.05)


# y_true doesn't have the class 3 but y_pred does
def test_classifier_scores_multiclass_classnotpresent_ytrue():
    train = pd.DataFrame({
        'x1': [100, 100, 100, 200, 200, 300, 500],
        'x2': [1, 2, 3, 1, 4, 5, 11],
        'x3': [4, 8, 12, 3, 12, 9, 22],
        'y': [1, 1, 1, 2, 2, 3, 4]
    })

    test = pd.DataFrame({
        'x1': [200, 300, 200, 500],
        'x2': [4, 7, 5, 10],
        'x3': [16, 11, 15, 20],
        'y': [1, 2, 2, 4]
    })

    x_train = train[['x1', 'x2', 'x3']]
    y_train = train['y']
    x_test = test[['x1', 'x2', 'x3']]
    y_test = test['y']

    clf = RidgeClassifier()
    clf.fit(x_train, y_train)
    y_pred = clf.predict(x_test)
    assert (y_pred == np.array([2, 3, 1, 4])).all()
    assert (np.array(y_test) == np.array([1, 2, 2, 4])).all()

    metrics_dict = classifier_scores(x_train=np.array(x_train),
                                     y_train=np.array(y_train),
                                     x_test=np.array(x_test),
                                     y_test=np.array(y_test), clf=clf)

    assert np.isclose(metrics_dict['roc_auc'], 0.541, rtol=0.05)


# y_true has class 4 but y_pred don't
# and y_pred has class 3 but y_test does
def test_classifier_scores_multiclass_classnotpresent_ytrue_y_pred():
    train = pd.DataFrame({
        'x1': ['100', '100', '100', '200', '200', '300'],
        'x2': [1, 2, 3, 1, 4, 5],
        'x3': [4, 8, 12, 3, 12, 9],
        'y': [1, 1, 1, 2, 2, 3]
    })

    test = pd.DataFrame({
        'x1': ['100', '200', '200'],
        'x2': [4, 2, 5],
        'x3': [16, 6, 13],
        'y': [1, 2, 4]
    })

    x_train = train[['x1', 'x2', 'x3']]
    y_train = train['y']
    x_test = test[['x1', 'x2', 'x3']]
    y_test = test['y']

    clf = RidgeClassifier()
    clf.fit(x_train, y_train)
    y_pred = clf.predict(x_test)
    assert (y_pred == np.array([1, 2, 3])).all()
    assert (np.array(y_test) == np.array([1, 2, 4])).all()

    metrics_dict = classifier_scores(x_train=np.array(x_train),
                                     y_train=np.array(y_train),
                                     x_test=np.array(x_test),
                                     y_test=np.array(y_test), clf=clf)

    assert np.isclose(metrics_dict['roc_auc'], 1.0, rtol=0.05)


def test_accuracy():
    y_true = [0, 1, 1, 1, 0, 1]
    y_pred = [0, 1, 0, 0, 1, 1]
    acc = Accuracy()
    assert acc(y_true, y_true) == 1
    assert acc(y_true, y_pred) == 0.5


def test_prediction_scores_from_df():
    train = pd.DataFrame({
        'c1': [100, 100, 100, 200, 200, 300, 500],
        'c2': [1, 2, 3, 1, 4, 5, 11],
        'c3': [4, 8, 12, 3, 12, 9, 22],
        'c4': [0, 1, 1, 0, 1, 0, 0]
    })

    test = pd.DataFrame({
        'c1': [200, 300, 200, 500],
        'c2': [4, 7, 5, 10],
        'c3': [16, 11, 15, 20],
        'c4': [1, 0, 1, 0]
    })

    metrics_dict = classifier_scores_from_df(df_train=train,
                                             df_test=test,
                                             target='c4',
                                             clf=RidgeClassifier())

    assert 'roc_auc' in metrics_dict

    metrics_dict = regressor_scores_from_df(df_train=train,
                                            df_test=test,
                                            target='c2',
                                            rgr=Ridge())

    assert 'r2_score' in metrics_dict


def test_modelling_metrics():
    # regression
    target = [1, 2, 3, 4, 4, 6, 9, 1]
    r2_score = R2Score()
    assert r2_score(target, target) == 1

    mse = MeanSquaredError()
    assert mse(target, target) == 0

    mae = MeanAbsoluteError()
    assert mae(target, target) == 0

    # binary classification
    target = [1, 0, 1, 1, 1, 0, 0, 1]
    precision = Precision()
    assert precision(target, target) == 1

    recall = Recall()
    assert recall(target, target) == 1

    f1score = F1Score()
    assert f1score(target, target) == 1

    train = pd.DataFrame({
        'x1': ['100', '100', '100', '200', '200', '300', '500'],
        'x2': [1, 2, 3, 1, 4, 5, 11],
        'x3': [4, 8, 12, 3, 12, 9, 4],
        'y': [0, 1, 1, 1, 0, 0, 0]
    })

    test = pd.DataFrame({
        'x1': ['100', '200', '200'],
        'x2': [4, 2, 5],
        'x3': [16, 6, 15],
        'y': [1, 0, 1]
    })

    x_train = train[['x1', 'x2', 'x3']]
    y_train = train['y']
    x_test = test[['x1', 'x2', 'x3']]
    y_test = test['y']

    clf = LogisticRegression()
    clf.fit(x_train, y_train)
    y_pred = clf.predict(x_test)
    y_pred_proba = clf.predict_proba(x_test)

    roc_curve = ROCCurve()
    assert roc_curve(y_test, y_pred_proba=y_pred_proba[:, 1].reshape(-1, 1)) is not None

    pr_curve = PRCurve()
    assert pr_curve(y_test, y_pred_proba=y_pred_proba[:, 1].reshape(-1, 1)) is not None

    cf_matrix = ConfusionMatrix(True)
    assert cf_matrix(y_test, y_pred, y_pred_proba) is not None

    # multiclass classification
    target = ['a', 'c', 'a', 'b', 'b', 'b', 'c', 'a']
    precision = Precision(multiclass=True)
    assert precision(target, target) == 1

    recall = Recall(multiclass=True)
    assert recall(target, target) == 1

    f1score = F1Score(multiclass=True)
    assert f1score(target, target) == 1

    train = pd.DataFrame({
        'x1': ['100', '100', '100', '200', '200', '300', '500'],
        'x2': [1, 2, 3, 1, 4, 5, 11],
        'x3': [4, 8, 12, 3, 12, 9, 4],
        'y': [1, 1, 1, 2, 2, 3, 1]
    })

    test = pd.DataFrame({
        'x1': ['100', '200', '200'],
        'x2': [4, 2, 5],
        'x3': [16, 6, 15],
        'y': [2, 3, 1]
    })

    x_train = train[['x1', 'x2', 'x3']]
    y_train = train['y']
    x_test = test[['x1', 'x2', 'x3']]
    y_test = test['y']

    clf = LogisticRegression()
    clf.fit(x_train, y_train)
    y_pred = clf.predict(x_test)
    y_pred_proba = clf.predict_proba(x_test)

    roc_curve = ROCCurve(True)
    assert roc_curve(y_test, y_pred, y_pred_proba) is None

    pr_curve = PRCurve(True)
    assert pr_curve(y_test, y_pred, y_pred_proba) is None

    cf_matrix = ConfusionMatrix(True)
    assert cf_matrix(y_test, y_pred, y_pred_proba) is not None



import copy

import numpy as np
import pandas as pd
import pytest

from synthesized_insight.metrics import privacy_cap_scorer
from synthesized_insight.metrics.privacy import AttributeInferenceAttackCAP, AttributeInferenceAttackML


@pytest.fixture(scope='module')
def df():
    df = pd.read_csv('tests/datasets/mini_credit.csv')
    return df


def check_output(metrics_cls, model, predictors, sensitive_col, df):
    privacy_metrics = metrics_cls(copy.deepcopy(model), sensitive_col, predictors)
    privacy_score_same = privacy_metrics(df, df)

    privacy_metrics = metrics_cls(copy.deepcopy(model), sensitive_col, predictors)
    new_df = df.copy()
    new_df[sensitive_col] = np.random.permutation(new_df[sensitive_col].values)
    privacy_score_permuted = privacy_metrics(new_df, df)
    assert privacy_score_same < privacy_score_permuted


@pytest.mark.slow
@pytest.mark.parametrize(
    "predictors,sensitive_col",
    [
        pytest.param(['SeriousDlqin2yrs', 'RevolvingUtilizationOfUnsecuredLines', 'DebtRatio'], 'age'),
        pytest.param(['age', 'NumberOfDependents', 'SeriousDlqin2yrs'], 'MonthlyIncome'),
        pytest.param(None, 'MonthlyIncome')
    ])
def test_privacy_metrics_numerical(df, predictors, sensitive_col):
    np.random.seed(6235901)
    model = 'Linear'
    check_output(AttributeInferenceAttackML, model, predictors, sensitive_col, df.head(1000).copy())


@pytest.mark.slow
@pytest.mark.parametrize(
    "predictors,sensitive_col",
    [
        pytest.param(['SeriousDlqin2yrs', 'RevolvingUtilizationOfUnsecuredLines', 'DebtRatio', 'age'], 'NumberOfDependents'),
        pytest.param(['NumberOfDependents', 'RevolvingUtilizationOfUnsecuredLines', 'DebtRatio', 'age'], 'SeriousDlqin2yrs'),
        pytest.param(None, 'SeriousDlqin2yrs')
    ])
def test_privacy_metrics_categorical_ml(df, predictors, sensitive_col):
    np.random.seed(6235901)
    model = 'RandomForest'
    check_output(AttributeInferenceAttackML, model, predictors, sensitive_col, df.head(1000).copy())


@pytest.mark.parametrize(
    "privacy_cap_model,predictors,sensitive_col",
    [
        pytest.param('GeneralizedCAP', ['SeriousDlqin2yrs', 'RevolvingUtilizationOfUnsecuredLines', 'DebtRatio', 'age'], 'NumberOfDependents'),
        pytest.param('DistanceCAP', ['SeriousDlqin2yrs', 'RevolvingUtilizationOfUnsecuredLines', 'DebtRatio', 'age'], 'NumberOfDependents'),
        pytest.param('GeneralizedCAP', ['SeriousDlqin2yrs', 'age'], 'NumberOfDependents'),
        pytest.param('DistanceCAP', ['SeriousDlqin2yrs', 'age'], 'NumberOfDependents'),
        pytest.param('GeneralizedCAP', None, 'NumberOfDependents')
    ])
def test_privacy_metrics_categorical_cap(df, privacy_cap_model, predictors, sensitive_col):
    np.random.seed(6235901)
    check_output(AttributeInferenceAttackCAP, privacy_cap_model, predictors, sensitive_col, df.head(1000).copy())


def test_privacy_score_ml_categorical(df):
    df = df.head(100)
    predictors = ['NumberOfDependents', 'age']
    sensitive_col = 'SeriousDlqin2yrs'
    ml_eval = AttributeInferenceAttackML('RandomForest', sensitive_col, predictors)
    assert ml_eval._privacy_score_categorical(df[sensitive_col], df[sensitive_col]) == 0

    permuted_sensitive_col = np.random.permutation(df[sensitive_col].values)
    assert ml_eval._privacy_score_categorical(df[sensitive_col], permuted_sensitive_col) > 0


def test_privacy_score_ml_numerical(df):
    df = df.head(100)
    predictors = ['MonthlyIncome', 'NumberOfDependents']
    sensitive_col = 'RevolvingUtilizationOfUnsecuredLines'
    ml_eval = AttributeInferenceAttackML('Linear', sensitive_col, predictors)
    assert ml_eval._privacy_score_numerical(df[sensitive_col], df[sensitive_col]) == 0

    permuted_sensitive_col = np.random.permutation(df[sensitive_col].values)
    assert ml_eval._privacy_score_numerical(df[sensitive_col], permuted_sensitive_col) > 0


def test_get_frequency():
    nums = [1, 1, 2, 3, 1, 2, 2, 1]
    elem = 1
    assert privacy_cap_scorer.get_frequency(nums, elem) == 0.5


def test_get_most_frequent_element():
    nums = [1, 1, 2, 2, 1]
    assert privacy_cap_scorer.get_most_frequent_element(nums) == 1


def test_get_hamming_distance():
    tpl1 = (1, 2, 3)
    assert privacy_cap_scorer.get_hamming_dist(tpl1, tpl1) == 0

    tpl2 = (1, 2, 4)
    assert privacy_cap_scorer.get_hamming_dist(tpl1, tpl2) == 1

    tpl2 = (5, 5, 5)
    assert privacy_cap_scorer.get_hamming_dist(tpl1, tpl2) == 3


def test_get_closest_neighbours():
    tpl_list = [(1, 2, 3), (2, 3, 4), (0, 1, 3), (0, 2, 4), (3, 2, 1)]
    target = (0, 2, 3)

    expected_closest_neighbours = [(1, 2, 3), (0, 1, 3), (0, 2, 4)]
    assert privacy_cap_scorer.get_closest_neighbours(tpl_list, target) == expected_closest_neighbours



"""This module contains various metrics used across synthesized."""
from typing import List, Optional, Sequence, Union, cast

import numpy as np
import pandas as pd
from scipy.spatial.distance import jensenshannon
from scipy.stats import entropy, wasserstein_distance
from sklearn.linear_model import LogisticRegression
from sklearn.preprocessing import OneHotEncoder

from ..check import Check, ColumnCheck
from ..modelling import ModellingPreprocessor
from .base import OneColumnMetric, TwoColumnMetric
from .utils import zipped_hist


class Mean(OneColumnMetric):
    name = "mean"

    @classmethod
    def check_column_types(cls, sr: pd.Series, check: Check = ColumnCheck()):
        if not check.affine(sr):
            return False
        return True

    def _compute_metric(self, sr: pd.Series):
        return affine_mean(sr)


class StandardDeviation(OneColumnMetric):
    name = "standard_deviation"

    def __init__(self, check: Check = ColumnCheck(), remove_outliers: float = 0.0):
        super().__init__(check)
        self.remove_outliers = remove_outliers

    @classmethod
    def check_column_types(cls, sr: pd.Series, check: Check = ColumnCheck()):
        if not check.affine(sr):
            return False
        return True

    def _compute_metric(self, sr: pd.Series):
        values = np.sort(sr.values)
        values = values[int(len(sr) * self.remove_outliers):int(len(sr) * (1.0 - self.remove_outliers))]
        return affine_stddev(pd.Series(values, name=sr.name))


class CramersV(TwoColumnMetric):
    """Cramér's V correlation coefficient between nominal variables.

    The statistic ranges from 0 to 1, where a value of 0 indicates there is no association between the variables,
    and 1 indicates maximal association (i.e one variable is completely determined by the other).
    """
    name = "cramers_v"
    symmetric = True

    @classmethod
    def check_column_types(cls, sr_a: pd.Series, sr_b: pd.Series, check: Check = ColumnCheck()):
        if not check.categorical(sr_a) or not check.categorical(sr_b):
            return False
        return True

    def _compute_metric(self, sr_a: pd.Series, sr_b: pd.Series):
        if len(sr_a.value_counts()) == 1 or len(sr_b.value_counts()) == 1:
            return 0

        table_orig = pd.crosstab(sr_a.astype(str), sr_b.astype(str))
        table = np.asarray(table_orig, dtype=np.float64)

        if table.min() == 0:
            table[table == 0] = 0.5

        n = table.sum()
        row = table.sum(1) / n
        col = table.sum(0) / n

        row = pd.Series(data=row, index=table_orig.index)
        col = pd.Series(data=col, index=table_orig.columns)

        itab = np.outer(row, col)
        probs = pd.DataFrame(
            data=itab, index=table_orig.index, columns=table_orig.columns
        )

        fit = table.sum() * probs
        expected = fit.to_numpy()

        real = table
        r, c = real.shape
        n = np.sum(real)
        v = np.sum((real - expected) ** 2 / (expected * n * min(r - 1, c - 1))) ** 0.5

        return v


class R2Mcfadden(TwoColumnMetric):
    """R2 Mcfadden correlation coefficient between catgorical and numerical variables.

    It trains two multinomial logistic regression models on the data, one using the numerical
    series as the feature and the other only using the intercept term as the input.
    The categorical column is used for the target labels. It then calculates the null
    and the model likelihoods based on them, which are used to compute the pseudo-R2 McFadden score,
    which is used as a correlation coefficient.
    The statistic ranges from 0 to 1, where a value of 0 indicates there is no association between the variables,
    and 1 indicates maximal association (i.e one variable is completely determined by the other).
    """
    name = "r2_mcfadden"

    @classmethod
    def check_column_types(cls, sr_a: pd.Series, sr_b: pd.Series, check: Check = ColumnCheck()):
        if not check.categorical(sr_a) or not check.continuous(sr_b):
            return False
        return True

    def _logistic_regression_r2(self,
                                df: pd.DataFrame,
                                y_label: str,
                                x_labels: List[str],
                                max_sample_size: int = 10_000) -> Union[None, float]:

        if len(x_labels) == 0:
            return None

        df = df[x_labels + [y_label]].dropna()
        df = pd.DataFrame(df.sample(min(max_sample_size, len(df))))  # explicit assignment for mypy

        if df[y_label].nunique() < 2:
            return None

        df_pre = ModellingPreprocessor.preprocess(df=df, target=y_label)
        x_labels_pre = list(filter(lambda v: v != y_label, df_pre.columns))

        x_array = df_pre[x_labels_pre].to_numpy()
        y_array = df_pre[y_label].to_numpy()

        rg = LogisticRegression()
        rg.fit(x_array, y_array)

        labels = df_pre[y_label].map({c: n for n, c in enumerate(rg.classes_)}).to_numpy()
        oh_labels = OneHotEncoder(sparse=False).fit_transform(labels.reshape(-1, 1))

        lp = rg.predict_log_proba(x_array)
        llf = np.sum(oh_labels * lp)

        rg = LogisticRegression()
        rg.fit(np.ones_like(y_array).reshape(-1, 1), y_array)

        lp = rg.predict_log_proba(x_array)
        llnull = np.sum(oh_labels * lp)

        psuedo_r2 = 1 - (llf / llnull)

        return psuedo_r2

    def _compute_metric(self, sr_a: pd.Series, sr_b: pd.Series):
        """Calculate the metric.
        Args:
            sr_a (pd.Series): values of a nominal variable.
            sr_b (pd.Series): values of numerical variable.
        Returns:
            The R2 Mcfadden correlation coefficient between sr_a and sr_b.
        """
        df = pd.DataFrame(data={sr_a.name: sr_a, sr_b.name: sr_b})
        r2 = self._logistic_regression_r2(df, y_label=str(sr_a.name), x_labels=[str(sr_b.name)])

        return r2


class EarthMoversDistance(TwoColumnMetric):
    """Earth mover's distance (aka 1-Wasserstein distance) between two nominal variables.

    The statistic ranges from 0 to 1, where a value of 0 indicates the two variables follow identical distributions,
    and a value of 1 indicates they follow completely different distributions.
    """
    name = "earth_movers_distance"

    @classmethod
    def check_column_types(cls, sr_a: pd.Series, sr_b: pd.Series, check: Check = ColumnCheck()) -> bool:
        if not check.categorical(sr_a) or not check.categorical(sr_b):
            return False
        return True

    def _compute_metric(self, sr_a: pd.Series, sr_b: pd.Series):
        """Calculate the metric.
        Args:
            sr_a (pd.Series): values of a nominal variable.
            sr_b (pd.Series): values of another nominal variable to compare.
        Returns:
            The earth mover's distance between sr_a and sr_b.
        """
        old = sr_a.to_numpy().astype(str)
        new = sr_b.to_numpy().astype(str)

        space = set(old).union(set(new))
        if len(space) > 1e4:
            return np.nan

        old_unique, counts = np.unique(old, return_counts=True)
        old_counts = dict(zip(old_unique, counts))

        new_unique, counts = np.unique(new, return_counts=True)
        new_counts = dict(zip(new_unique, counts))

        p = np.array([float(old_counts[x]) if x in old_counts else 0.0 for x in space])
        q = np.array([float(new_counts[x]) if x in new_counts else 0.0 for x in space])

        p /= np.sum(p)
        q /= np.sum(q)
        distance = 0.5 * np.sum(np.abs(p.astype(np.float64) - q.astype(np.float64)))
        return distance


class KullbackLeiblerDivergence(TwoColumnMetric):
    """Kullback–Leibler Divergence or Relative Entropy between two probability distributions.

    The statistic ranges from 0 to 1, where a value of 0 indicates the two variables follow identical distributions,
    and a value of 1 indicates they follow completely different distributions.
    """
    name = "kullback_leibler_divergence"

    @classmethod
    def check_column_types(cls, sr_a: pd.Series, sr_b: pd.Series, check: Check = ColumnCheck()) -> bool:
        if check.continuous(sr_a) and check.continuous(sr_b):
            return True
        if check.categorical(sr_a) and check.categorical(sr_b):
            return True
        return False

    def _compute_metric(self, sr_a: pd.Series, sr_b: pd.Series):
        """Calculate the metric.
        Args:
            sr_a (pd.Series): values of a variable.
            sr_b (pd.Series): values of another variable to compare.
        Returns:
            The kullback-leibler divergence between sr_a and sr_b.
        """
        (p, q) = zipped_hist((sr_a, sr_b), check=self.check)
        return entropy(np.array(p), np.array(q))


class JensenShannonDivergence(TwoColumnMetric):
    """Jensen-Shannon Divergence between two probability distributions.

    The statistic ranges from 0 to 1, where a value of 0 indicates the two variables follow identical distributions,
    and a value of 1 indicates they follow completely different distributions.
    """
    name = "jensen_shannon_divergence"

    @classmethod
    def check_column_types(cls, sr_a: pd.Series, sr_b: pd.Series, check: Check = ColumnCheck()) -> bool:
        if check.continuous(sr_a) and check.continuous(sr_b):
            return True
        if check.categorical(sr_a) and check.categorical(sr_b):
            return True
        return False

    def _compute_metric(self, sr_a: pd.Series, sr_b: pd.Series):
        """Calculate the metric.
        Args:
            sr_a (pd.Series): values of a variable.
            sr_b (pd.Series): values of another variable to compare.
        Returns:
            The jensen-shannon divergence between sr_a and sr_b.
        """
        (p, q) = zipped_hist((sr_a, sr_b), check=self.check)
        return jensenshannon(p, q)


class HellingerDistance(TwoColumnMetric):
    """Hellinger Distance between two probability distributions.

    The statistic ranges from 0 to 1, where a value of 0 indicates the two variables follow identical distributions,
    and a value of 1 indicates they follow completely different distributions.
    """
    name = "hellinger_distance"

    @classmethod
    def check_column_types(cls, sr_a: pd.Series, sr_b: pd.Series, check: Check = ColumnCheck()) -> bool:
        if check.continuous(sr_a) and check.continuous(sr_b):
            return True
        if check.categorical(sr_a) and check.categorical(sr_b):
            return True
        return False

    def _compute_metric(self, sr_a: pd.Series, sr_b: pd.Series):
        """Calculate the metric.
        Args:
            sr_a (pd.Series): values of a variable.
            sr_b (pd.Series): values of another variable to compare.
        Returns:
            The hellinger distance between sr_a and sr_b.
        """
        (p, q) = zipped_hist((sr_a, sr_b), check=self.check)
        return np.linalg.norm(np.sqrt(cast(pd.Series, p)) - np.sqrt(cast(pd.Series, q))) / np.sqrt(2)


class Norm(TwoColumnMetric):
    """Norm between two probability distributions.

    The statistic ranges from 0 to 1, where a value of 0 indicates the two variables follow identical distributions,
    and a value of 1 indicates they follow completely different distributions.

    Args:
        ord (Union[str, int], optional):
                The order of the norm. Possible values include positive numbers, 'fro', 'nuc'.
                See numpy.linalg.norm for more details. Defaults to 2.
    """
    name = "norm"

    def __init__(self, check: Check = ColumnCheck(), ord: Union[None, float, str] = 2.0):
        super().__init__(check)
        self.ord = ord

    @classmethod
    def check_column_types(cls, sr_a: pd.Series, sr_b: pd.Series, check: Check = ColumnCheck()) -> bool:
        if check.continuous(sr_a) and check.continuous(sr_b):
            return True
        if check.categorical(sr_a) and check.categorical(sr_b):
            return True
        return False

    def _compute_metric(self, sr_a: pd.Series, sr_b: pd.Series):
        """Calculate the metric.
        Args:
            sr_a (pd.Series): values of a variable.
            sr_b (pd.Series): values of another variable to compare.
        Returns:
            The lp-norm between sr_a and sr_b.
        """
        (p, q) = zipped_hist((sr_a, sr_b), check=self.check)
        if p is not None and q is not None:
            return np.linalg.norm(cast(pd.Series, p) - cast(pd.Series, q), ord=self.ord)  # type: ignore
        return None


class EarthMoversDistanceBinned(TwoColumnMetric):
    """
    Earth movers distance (EMD), aka Wasserstein 1-distance, between two histograms.

    The histograms can represent counts of nominal categories or counts on
    an ordinal range. If the latter, they must have equal binning.

    Args:
        bin_edges: Optional; If given, this must be an iterable of bin edges for x and y,
                i.e the output of np.histogram_bin_edges. If None, then it is assumed
                that the data represent counts of nominal categories, with no meaningful
                distance between bins.

    """
    name = "earth_movers_distance_binned"

    def __init__(self,
                 check: Check = ColumnCheck(),
                 bin_edges: Optional[Union[pd.Series, Sequence, np.ndarray]] = None):
        super().__init__(check)
        self.bin_edges = bin_edges

    @classmethod
    def check_column_types(cls, sr_a: pd.Series, sr_b: pd.Series, check: Check = ColumnCheck()) -> bool:
        # Histograms can appear to be continuous even if they are categorical in nature
        return True

    def _compute_metric(self, sr_a: pd.Series, sr_b: pd.Series):
        """
        Calculate the EMD between two binned continuous data or two 1d histograms.

        Histograms must have an equal number of bins. They are not required to be normalised,
        and distances between bins are measured using a Euclidean metric.

        Returns:
            The earth mover's distance.
        """
        if sr_a.sum() == 0 and sr_b.sum() == 0:
            return 0.
        elif sr_a.sum() == 0 or sr_b.sum() == 0:
            return 1.

        # normalise counts for consistency with scipy.stats.wasserstein
        with np.errstate(divide='ignore', invalid='ignore'):
            x = np.nan_to_num(sr_a / sr_a.sum())
            y = np.nan_to_num(sr_b / sr_b.sum())

        if self.bin_edges is None:
            # if bins not given, histograms are assumed to be counts of nominal categories,
            # and therefore distances betwen bins are meaningless. Set to all distances to
            # unity to model this.
            distance = 0.5 * np.sum(np.abs(x.astype(np.float64) - y.astype(np.float64)))
        else:
            # otherwise, use pair-wise euclidean distances between bin centers for scale data
            bin_centers = self.bin_edges[:-1] + np.diff(self.bin_edges) / 2.
            distance = wasserstein_distance(bin_centers, bin_centers, u_weights=x, v_weights=y)
        return distance


def affine_mean(sr: pd.Series):
    """function for calculating means of affine values"""
    mean = np.nanmean(sr.values - np.array(0, dtype=sr.dtype))
    return mean + np.array(0, dtype=sr.dtype)


def affine_stddev(sr: pd.Series):
    """function for calculating standard deviations of affine values"""
    d = sr - affine_mean(sr)
    u = d / np.array(1, dtype=d.dtype)
    s = np.sqrt(np.sum(u**2))
    return s * np.array(1, dtype=d.dtype)



from .base import OneColumnMetric, TwoColumnMetric, TwoColumnTest, TwoDataFrameMetric
from .metrics import (
    CramersV,
    EarthMoversDistance,
    EarthMoversDistanceBinned,
    HellingerDistance,
    JensenShannonDivergence,
    KullbackLeiblerDivergence,
    Mean,
    Norm,
    R2Mcfadden,
    StandardDeviation,
)
from .metrics_usage import CorrMatrix, DiffCorrMatrix, TwoColumnMap
from .modelling_metrics import (
    ROCAUC,
    Accuracy,
    ConfusionMatrix,
    F1Score,
    MeanAbsoluteError,
    MeanSquaredError,
    PRCurve,
    Precision,
    PredictiveModellingScore,
    R2Score,
    Recall,
    ROCCurve,
)
from .statistical_tests import (
    BinomialDistanceTest,
    KendallTauCorrelationTest,
    KolmogorovSmirnovDistanceTest,
    KruskalWallisTest,
    SpearmanRhoCorrelationTest,
)

__all__ = ['OneColumnMetric', 'TwoColumnMetric', 'TwoColumnMap', 'CorrMatrix', 'TwoColumnTest',
           'DiffCorrMatrix', 'CramersV', 'EarthMoversDistance', 'KendallTauCorrelationTest',
           'KolmogorovSmirnovDistanceTest', 'Mean', 'SpearmanRhoCorrelationTest', 'StandardDeviation',
           'ROCAUC', 'Accuracy', 'ConfusionMatrix', 'F1Score', 'MeanAbsoluteError', 'KruskalWallisTest',
           'MeanSquaredError', 'PRCurve', 'Precision', 'R2Score', 'Recall', 'ROCCurve', 'Norm',
           'PredictiveModellingScore', 'TwoDataFrameMetric', 'EarthMoversDistanceBinned', 'JensenShannonDivergence',
           'KullbackLeiblerDivergence', 'HellingerDistance', 'BinomialDistanceTest', 'R2Mcfadden']



import numpy as np
import pandas as pd
import pytest

from synthesized_insight.check import ColumnCheck
from synthesized_insight.metrics import (
    CorrMatrix,
    CramersV,
    DiffCorrMatrix,
    EarthMoversDistance,
    KolmogorovSmirnovDistanceTest,
    KruskalWallisTest,
    TwoColumnMap,
)


@pytest.fixture(scope='module')
def data():
    df = pd.read_csv('tests/datasets/mini_credit.csv').dropna().reset_index(drop=True)
    categorical_cols = []
    continuous_cols = []

    check = ColumnCheck()
    for col in df.columns:
        if check.continuous(df[col]):
            continuous_cols.append(col)
        else:
            categorical_cols.append(col)

    return df, categorical_cols, continuous_cols


def test_two_column_map(data):
    df, categorical_cols, continuous_cols = data[0], data[1], data[2]
    df1 = df.sample(1000).reset_index(drop=True)
    df2 = df.sample(1000).reset_index(drop=True)

    emd = EarthMoversDistance()
    ksdt = KolmogorovSmirnovDistanceTest()

    col_map = TwoColumnMap(emd)
    emd_map_df = col_map(df1, df2)
    assert col_map.name == f'{str(emd)}_map'

    assert set(emd_map_df.columns.to_list()) == set(['metric_val'])
    assert all(not np.isnan(emd_map_df['metric_val'][cat]) for cat in categorical_cols)
    assert all(np.isnan(emd_map_df['metric_val'][cont]) for cont in continuous_cols)

    col_map = TwoColumnMap(ksdt)
    kdst_map_df = col_map(df1, df2)
    assert col_map.name == f'{str(ksdt)}_map'
    assert set(kdst_map_df.columns.to_list()) == set(['metric_val', 'metric_pval'])
    assert all(np.isnan(kdst_map_df['metric_val'][cat]) for cat in categorical_cols)
    assert all(np.isnan(kdst_map_df['metric_pval'][cat]) for cat in categorical_cols)
    assert all(not np.isnan(kdst_map_df['metric_val'][cont]) for cont in continuous_cols)
    assert all(not np.isnan(kdst_map_df['metric_pval'][cont]) for cont in continuous_cols)


def test_metric_matrix(data):
    df, categorical_cols, continuous_cols = data[0], data[1], data[2]
    df1 = df.sample(1000).reset_index(drop=True)
    df2 = df.sample(1000).reset_index(drop=True)

    kwt = KruskalWallisTest()
    cmt = CorrMatrix(kwt)
    assert cmt.name == f'{str(kwt)}_matrix'
    kwt_val_df, kwt_pval_df = cmt(df)
    assert kwt_pval_df is not None
    assert all(not np.isnan(kwt_val_df[cont1][cont2]) and not np.isnan(kwt_val_df[cont2][cont1]) for cont1 in continuous_cols for cont2 in continuous_cols if cont1 != cont2)
    assert all(np.isnan(kwt_val_df[cat][cont]) and np.isnan(kwt_val_df[cont][cat]) for cat in categorical_cols for cont in continuous_cols)
    assert all(np.isnan(kwt_val_df[cat1][cat2]) and np.isnan(kwt_val_df[cat2][cat1]) for cat1 in categorical_cols for cat2 in categorical_cols)

    kwt_diff_mat = DiffCorrMatrix(kwt)
    diff = kwt_diff_mat(df1, df2)
    assert kwt_diff_mat.name == f'diff_{str(kwt)}'
    assert all(not np.isnan(diff[cont1][cont2]) and not np.isnan(diff[cont2][cont1]) for cont1 in continuous_cols for cont2 in continuous_cols if cont1 != cont2)
    assert all(np.isnan(diff[cat][cont]) and np.isnan(diff[cont][cat]) for cat in categorical_cols for cont in continuous_cols)
    assert all(np.isnan(diff[cat1][cat2]) and np.isnan(diff[cat2][cat1]) for cat1 in categorical_cols for cat2 in categorical_cols)

    cmv = CramersV()
    cmt = CorrMatrix(cmv)
    assert cmt.name == f'{str(cmv)}_matrix'
    cmv_val_df, cmv_pval_df = cmt(df)
    assert cmv_pval_df is None
    assert all(np.isnan(cmv_val_df[cont1][cont2]) and np.isnan(cmv_val_df[cont2][cont1]) for cont1 in continuous_cols for cont2 in continuous_cols)
    assert all(np.isnan(cmv_val_df[cat][cont]) and np.isnan(cmv_val_df[cont][cat]) for cat in categorical_cols for cont in continuous_cols)
    assert all(not np.isnan(cmv_val_df[cat1][cat2]) and not np.isnan(cmv_val_df[cat2][cat1]) for cat1 in categorical_cols for cat2 in categorical_cols if cat1 != cat2)

    cmv_diff_mat = DiffCorrMatrix(cmv)
    diff = cmv_diff_mat(df1, df2)
    assert cmv_diff_mat.name == f'diff_{str(cmv)}'
    assert all(np.isnan(diff[cont1][cont2]) and np.isnan(diff[cont2][cont1]) for cont1 in continuous_cols for cont2 in continuous_cols)
    assert all(np.isnan(diff[cat][cont]) and np.isnan(diff[cont][cat]) for cat in categorical_cols for cont in continuous_cols)
    assert all(not np.isnan(diff[cat1][cat2]) and not np.isnan(diff[cat2][cat1]) for cat1 in categorical_cols for cat2 in categorical_cols if cat1 != cat2)



from abc import ABC, abstractmethod
from collections import defaultdict
from typing import Any, Dict, List, Tuple

import numpy as np
import pandas as pd


class CAPScorer(ABC):
    """Calculates the privacy score using CAP (Correct Attribution Probability) model

    Fits the model by mapping the predictors keys to the sensitive values using the synthetic data.
    These mappings are stored in predictors_to_sensitive_val_map dictionary.
    Then, evaluates the privacy of the original dataset using specific implementation of score(..) method
    in the child classes.
    """

    def __init__(self) -> None:
        self.predictors_to_sensitive_val_map: Dict[Tuple, List[Any]] = {}

    def fit(self, predictors_data: pd.DataFrame, sensitive_data: pd.Series):
        """Fit the privacy scorer on the synthetic data.

        Go through each row of the predictors_data and map the tuple of the predictor values to
        corresponding sensitive column value. Tuple of the predictor values is the key for the dictionary.
        If multiple rows contain same predictor tuple, then add their corresponding sensitive atribute value
        to the list of sensitive values corresponding to the key.

        Args:
            predictors_data: Values of the predictor columns
            sensitive_data: Values of the sensitive column

        Returns:
            self
        """
        for (pval, sval) in zip(predictors_data.values, sensitive_data.values):
            if sval == np.nan:
                continue

            key = tuple(pval)
            if key in self.predictors_to_sensitive_val_map:
                self.predictors_to_sensitive_val_map[key].append(sval)
            else:
                self.predictors_to_sensitive_val_map[key] = [sval]
        return self

    @abstractmethod
    def score(self, predictors_data: pd.DataFrame, true_senstive_vals: pd.Series) -> float:
        """Score the privacy of the sensitive column of the original df using synthetic data

        Args:
            predictors_data: Values of predictor columns in original df
            true_senstive_vals: Values of the sensitive column in original df

        Returns:
            Privacy score between 0 and 1, 0 means no privacy and 1 means absolute privacy
        """
        pass


def get_frequency(samples, element) -> float:
    """Find the frequency of the given element in given samples

    Args:
        samples: List of elements
        element: Element whose frequency needs to be determined

    Returns:
        Frequency of the element
    """
    return samples.count(element) / len(samples)


def get_most_frequent_element(elements):
    """Find the element with the highest frequency

    Args:
        elements: List of element

    Returns:
        The element with highest frequency
    """
    freq_map = defaultdict(int)
    most_freq_elem = None
    highest_freq = 0
    for elem in elements:
        if elem is None:
            continue

        freq = freq_map[elem] + 1
        freq_map[elem] = freq
        if freq > highest_freq:
            highest_freq = freq
            most_freq_elem = elem

    return most_freq_elem


def get_hamming_dist(tuple1: Tuple[Any], tuple2: Tuple[Any]) -> int:
    """Find the hamming distance between two tuples.

    The hamming distance is the number of respective elements in the tuples that are not equal.
    E.g. tuple1 = (1, 23, 'a'), tuple2 = (1, 1, 'a')
    Hamming distance for the above example is 1,
    since the 2nd element of the two tuples don't match.

    Args:
        tuple1: First tuple
        tuple2: Second tuple

    Returns:
        Number of respective elements that don't match
    """
    dist = 0
    assert len(tuple1) == len(tuple2), ('Tuples must have the same length')

    for val1, val2 in zip(tuple1, tuple2):
        if val1 != val2:
            dist += 1

    return dist


def get_closest_neighbours(samples, target) -> List[Any]:
    """Find elements in a given list that are closest to the given target using hamming distance.

    Arguments:
        samples: The list from which to select neighbours to the given target.
        target: The target element

    Returns:
        Elements in samples that are closest to the target.
    """
    dist = float('inf')
    closest_nei = []
    for sample in samples:
        hamming_dist = get_hamming_dist(target, sample)
        if hamming_dist < dist:
            dist = hamming_dist
            closest_nei = [sample, ]
        elif hamming_dist == dist:
            closest_nei.append(sample)

    return closest_nei


class GeneralizedCAPScorer(CAPScorer):
    """Calculates the privacy score using CAP (Correct Attribution Probability) model using
    exact match of the original df key in the synthetic table

    Fits the model by mapping the predictors keys to the sensitive values using the synthetic data.
    Then, evaluates the privacy of the original dataset by trying to find the sensitive attribute of the
    original data in the dictionary corresponding to the original df predictor key.
    """

    def __init__(self) -> None:
        super().__init__()

    def score(self, orig_predictors: pd.DataFrame, orig_sensitive: pd.Series) -> float:
        """Gives the privacy score based on evaluation of the original data using synthetic data

        For each row of predictors_data, calculate as follows:
        find the list of sensitive values corresponding to predictor key in the fitted
        predictors_to_sensitive_val_map dictionary, and then find the frequency of true sensitive
        value in this list. Add this frequency to the score.

        Average out the score to get the final privacy score.

        Args:
            orig_predictors: Values of predictor columns in original df
            orig_sensitive: Values of the sensitive column in original df

        Returns:
            Privacy score between 0 and 1, 0 means no privacy and 1 means absolute privacy
        """

        score = 0.0
        count = 0

        for (pval, sval) in zip(orig_predictors.values, orig_sensitive.values):
            key = tuple(pval)
            if key in self.predictors_to_sensitive_val_map:
                score += get_frequency(self.predictors_to_sensitive_val_map[key], sval)
                count += 1

        if count == 0:
            return 0

        return 1.0 - score / count


class DistanceCAPScorer(CAPScorer):
    """Calculates the privacy score using CAP (Correct Attribution Probability) distance model
    which uses neighbouring keys of the original df key to evaluate privacy

    Fits the model by mapping the predictors keys to the sensitive values using the synthetic data.
    Then, evaluates the privacy of the original dataset by trying to find the sensitive attribute of the
    original data in the dictionary corresponding to the neighbouring keys of the original df predictor key.
    """
    def __init__(self) -> None:
        super().__init__()

    def score(self, orig_predictors: pd.DataFrame, orig_sensitive: pd.Series) -> float:
        """Gives the privacy score based on evaluation of the original data using synthetic data

        For each row of predictors_data, calculate as follows:
        find the closest neighbouring keys of original df predictor key from predictors_to_sensitive_val_map,
        get the list of sensitive values corresponding to these neighbouring keys from
        predictors_to_sensitive_val_map dictionary, and then find the frequency of true sensitive
        value in this list. Add this frequency to the score.

        Average it out to get the final privacy score.

        Args:
            orig_predictors: Values of predictor columns in original df
            orig_sensitive: Values of the sensitive column in original df

        Returns:
            Privacy score between 0 and 1, 0 means no privacy and 1 means absolute privacy
        """
        score = 0.0
        count = 0

        for (pval, sval) in zip(orig_predictors.values, orig_sensitive.values):
            key = tuple(pval)

            closest_neighbours = get_closest_neighbours(self.predictors_to_sensitive_val_map.keys(), key)
            sensitive_vals = []
            for nei in closest_neighbours:
                sensitive_vals.extend(self.predictors_to_sensitive_val_map[nei])

            if sval in sensitive_vals:
                score += get_frequency(sensitive_vals, sval)
                count += 1

        if count == 0:
            return 0

        return 1.0 - score / count



import logging
from math import sqrt
from typing import Dict, List, Optional, Tuple, Type, Union, cast

import numpy as np
import pandas as pd
from sklearn.base import BaseEstimator, ClassifierMixin, RegressorMixin, clone
from sklearn.ensemble import (
    GradientBoostingClassifier,
    GradientBoostingRegressor,
    RandomForestClassifier,
    RandomForestRegressor,
)
from sklearn.linear_model import LogisticRegression, Ridge, RidgeClassifier
from sklearn.model_selection import train_test_split
from sklearn.neural_network import MLPClassifier, MLPRegressor
from sklearn.svm import LinearSVC, LinearSVR

logger = logging.getLogger(__name__)

"""A dictionary of sklearn classifiers with fit/predict methods."""
CLASSIFIERS: Dict[str, Type[ClassifierMixin]] = {
    'Linear': RidgeClassifier,
    'Logistic': LogisticRegression,
    'GradientBoosting': GradientBoostingClassifier,
    'RandomForest': RandomForestClassifier,
    'MLP': MLPClassifier,
    'LinearSVM': LinearSVC
}

"""A dictionary of sklearn regressors with fit/predict methods."""
REGRESSORS: Dict[str, Type[RegressorMixin]] = {
    'Linear': Ridge,
    'GradientBoosting': GradientBoostingRegressor,
    'RandomForest': RandomForestRegressor,
    'MLP': MLPRegressor,
    'LinearSVM': LinearSVR
}

RANDOM_SEED = 42


def sample_split_data(df: pd.DataFrame,
                      response_variable: str,
                      explanatory_variables: Optional[List[str]] = None,
                      test_size: Optional[float] = 0.2,
                      sample_size: Optional[int] = None
                      ) -> Tuple[pd.DataFrame, pd.DataFrame]:

    if explanatory_variables is not None:
        df = cast(pd.DataFrame, df[np.concatenate((explanatory_variables, [response_variable]))])

    if sample_size is not None and sample_size < len(df):
        df = cast(pd.DataFrame, df.sample(sample_size, random_state=RANDOM_SEED))

    if df[response_variable].nunique() > sqrt(len(df)):
        df_train, df_test = train_test_split(df,
                                             test_size=test_size,
                                             random_state=RANDOM_SEED)
    elif all(df[response_variable].value_counts().values > 1)\
            and df[response_variable].nunique() <= sqrt(len(df)):
        df_train, df_test = train_test_split(df,
                                             test_size=test_size,
                                             stratify=df[response_variable],
                                             random_state=RANDOM_SEED)
    else:
        df_train, df_test = train_test_split(df,
                                             test_size=test_size,
                                             random_state=RANDOM_SEED)

        train_unique = df_train[response_variable].unique()
        target_in_train = df_test[response_variable]\
            .apply(lambda y: True if y in train_unique else False)
        df_test = df_test[target_in_train]

    return df_train, df_test


def _get_regression_model(model: Union[str, RegressorMixin],
                          copy_model: bool) -> RegressorMixin:
    estimator = None
    if isinstance(model, str):
        if model in REGRESSORS:
            estimator = REGRESSORS[model]()
        else:
            raise KeyError(f"Selected model '{model}'\
                                 not available for REGRESSION.")
    elif isinstance(model, BaseEstimator):
        if not isinstance(model, RegressorMixin):
            raise KeyError("Given model is not available for REGRESSION.")

        if copy_model:
            estimator = clone(model)
        else:
            estimator = model
    else:
        raise ValueError("Given model type is not compatible")

    return estimator


def _get_classification_model(model: Union[str, ClassifierMixin],
                              copy_model: bool) -> ClassifierMixin:
    estimator = None
    if isinstance(model, str):
        if model in CLASSIFIERS:
            estimator = CLASSIFIERS[model]()
        else:
            raise KeyError(f"Selected model '{model}'\
                                 not available for CLASSIFICATION.")
    elif isinstance(model, BaseEstimator):
        if not isinstance(model, ClassifierMixin):
            raise KeyError("Given model is not available for CLASSIFICATION.")

        if copy_model:
            estimator = clone(model)
        else:
            estimator = model
    else:
        raise ValueError("Given model type is not compatible")
    return estimator


def check_model_type(model: Union[str, ClassifierMixin, RegressorMixin],
                     copy_model: bool,
                     task: str) -> Union[ClassifierMixin, RegressorMixin]:
    """Given a model (string or type) and task (classification or regreesion),
     return the estimator corresponding to that model"""
    if task == 'clf':
        estimator = _get_classification_model(model=model, copy_model=copy_model)
    elif task == 'rgr':
        estimator = _get_regression_model(model=model, copy_model=copy_model)
    else:
        raise ValueError(f"Given task '{task}' not recognized")

    return estimator



import pandas as pd
import pytest

from synthesized_insight.metrics import EarthMoversDistance, HellingerDistance
from synthesized_insight.metrics.confidence_interval import compute_binomial_interval, compute_bootstrap_interval


@pytest.mark.parametrize(
    'metric, data', [
        (EarthMoversDistance(), (pd.Series(['a', 'b', 'c']), pd.Series(['c', 'b', 'a']))),
        (HellingerDistance(), (pd.Series([1, 2, 3]), pd.Series([0, 0, 0])))
    ])
def test_bootstrap_interval(metric, data):
    conf_interval = compute_bootstrap_interval(metric, data[0], data[0])
    assert conf_interval.level == 0.95
    assert conf_interval.limits[0] is not None and conf_interval.limits[1] is not None

    conf_interval = compute_bootstrap_interval(metric, data[0], data[1])
    assert conf_interval.level == 0.95
    assert conf_interval.limits[0] is not None and conf_interval.limits[1] is not None


def test_binomial_interval():
    conf_interval = compute_binomial_interval(pd.Series([1, 1]), pd.Series([0, 0]), confidence_level=0.99)
    assert conf_interval.level == 0.99
    assert conf_interval.limits[0] is not None and conf_interval.limits[1] is not None

    conf_interval = compute_binomial_interval(pd.Series([1, 0]), pd.Series([1, 0]), confidence_level=0.80)
    assert conf_interval.level == 0.80
    assert conf_interval.limits[0] is not None and conf_interval.limits[1] is not None

    conf_interval = compute_binomial_interval(pd.Series([1, 0, 1, 1]), pd.Series([1, 0, 1, 0]))
    assert conf_interval.level == 0.95
    assert conf_interval.limits[0] is not None and conf_interval.limits[1] is not None



import logging
from typing import Any, Dict, List, Optional, Type, Union, cast

import numpy as np
import pandas as pd
from sklearn.base import BaseEstimator, ClassifierMixin, RegressorMixin
from sklearn.exceptions import NotFittedError
from sklearn.metrics import (
    accuracy_score,
    confusion_matrix,
    f1_score,
    mean_absolute_error,
    mean_squared_error,
    precision_recall_curve,
    precision_score,
    r2_score,
    recall_score,
    roc_auc_score,
    roc_curve,
)
from sklearn.preprocessing import OneHotEncoder, normalize
from sklearn.utils.validation import check_is_fitted

from ..check import Check, ColumnCheck
from ..modelling import ModellingPreprocessor, check_model_type, sample_split_data
from .base import ClassificationMetric, ClassificationPlotMetric, DataFrameMetric, RegressionMetric

logger = logging.getLogger(__name__)

MAX_ANALYSIS_SAMPLE_SIZE = 10000


class Accuracy(ClassificationMetric):
    name = "accuracy"

    def __call__(self, y_true: np.ndarray, y_pred: Optional[np.ndarray] = None,
                 y_pred_proba: Optional[np.ndarray] = None) -> Union[float]:
        assert y_pred is not None
        return accuracy_score(y_true, y_pred)


class Precision(ClassificationMetric):
    name = "precision"

    def __call__(self, y_true: np.ndarray, y_pred: Optional[np.ndarray] = None,
                 y_pred_proba: Optional[np.ndarray] = None) -> Union[float]:
        assert y_pred is not None
        if self.multiclass is False:
            return precision_score(y_true, y_pred)
        else:
            return precision_score(y_true, y_pred, average='micro')


class Recall(ClassificationMetric):
    name = "recall"

    def __call__(self,
                 y_true: np.ndarray,
                 y_pred: Optional[np.ndarray] = None,
                 y_pred_proba: Optional[np.ndarray] = None) -> Union[float]:
        assert y_pred is not None
        if self.multiclass is False:
            return recall_score(y_true, y_pred)
        else:
            return recall_score(y_true, y_pred, average='micro')


class F1Score(ClassificationMetric):
    name = "f1_score"

    def __call__(self,
                 y_true: np.ndarray,
                 y_pred: Optional[np.ndarray] = None,
                 y_pred_proba: Optional[np.ndarray] = None) -> float:
        assert y_pred is not None
        if self.multiclass is False:
            return f1_score(y_true, y_pred)
        else:
            return f1_score(y_true, y_pred, average='micro')


class ROCAUC(ClassificationMetric):
    name = "roc_auc"

    def __call__(self,
                 y_true: np.ndarray,
                 y_pred: Optional[np.ndarray] = None,
                 y_pred_proba: Optional[np.ndarray] = None) -> float:
        assert y_pred_proba is not None
        if len(np.unique(y_true)) < 2:
            return 1.0
        if self.multiclass is False:
            return roc_auc_score(y_true, y_pred_proba)
        else:
            return roc_auc_score(y_true, y_pred_proba, multi_class='ovo')


class ROCCurve(ClassificationPlotMetric):
    name = "roc_curve"

    def __init__(self, multiclass: bool = False):
        self.plot = True
        super(ROCCurve, self).__init__(multiclass=multiclass)

    def __call__(self,
                 y_true: np.ndarray,
                 y_pred: Optional[np.ndarray] = None,
                 y_pred_proba: Optional[np.ndarray] = None
                 ) -> Union[float, None]:
        assert y_pred_proba is not None
        if self.multiclass is False:
            return roc_curve(y_true, y_pred_proba)
        else:
            logger.warning("ROC Curve plot not available\
                            for multi-class classification.")
            return None


class PRCurve(ClassificationPlotMetric):
    name = "pr_curve"

    def __init__(self, multiclass: bool = False):
        self.plot = True
        super(PRCurve, self).__init__(multiclass=multiclass)

    def __call__(self,
                 y_true: np.ndarray,
                 y_pred: Optional[np.ndarray] = None,
                 y_pred_proba: Optional[np.ndarray] = None
                 ) -> Union[float, None]:
        assert y_pred_proba is not None
        if self.multiclass is False:
            return precision_recall_curve(y_true, y_pred_proba)
        else:
            logger.warning("PR Curve plot not available\
                            for multi-class classification.")
            return None


class ConfusionMatrix(ClassificationPlotMetric):
    name = "confusion_matrix"

    def __init__(self, multiclass: bool = False):
        self.plot = True
        super(ConfusionMatrix, self).__init__(multiclass=multiclass)

    def __call__(self,
                 y_true: np.ndarray,
                 y_pred: Optional[np.ndarray] = None,
                 y_pred_proba: Optional[np.ndarray] = None
                 ) -> Union[int, float, None]:
        assert y_pred is not None
        return confusion_matrix(y_true, y_pred)


class MeanAbsoluteError(RegressionMetric):
    name = "mean_absolute_error"

    def __init__(self):
        super(MeanAbsoluteError, self).__init__()

    def __call__(self,
                 y_true: np.ndarray = None,
                 y_pred: np.ndarray = None,
                 **kwargs) -> Union[float]:
        assert y_true is not None and y_pred is not None
        return mean_absolute_error(y_true, y_pred)


class MeanSquaredError(RegressionMetric):
    name = "mean_squared_error"

    def __init__(self):
        super(MeanSquaredError, self).__init__()

    def __call__(self,
                 y_true: np.ndarray = None,
                 y_pred: np.ndarray = None,
                 **kwargs) -> Union[float]:
        assert y_true is not None and y_pred is not None
        return mean_squared_error(y_true, y_pred)


class R2Score(RegressionMetric):
    name = "r2_score"

    def __init__(self):
        super(R2Score, self).__init__()

    def __call__(self,
                 y_true: np.ndarray = None,
                 y_pred: np.ndarray = None,
                 **kwargs) -> Union[float]:
        assert y_true is not None and y_pred is not None
        return r2_score(y_true, y_pred)


REGRESSION_METRICS: Dict[str, Type[RegressionMetric]] = {
    cast(str, m.name): m for m in [R2Score, MeanSquaredError,
                                   MeanAbsoluteError]
}
CLASSIFICATION_METRICS: Dict[str, Type[ClassificationMetric]] = {
    cast(str, m.name): m for m in [Accuracy, Precision, Recall,
                                   F1Score, ROCAUC]
}
CLASSIFICATION_PLOT_METRICS: Dict[str, Type[ClassificationPlotMetric]] = {
    cast(str, m.name): m for m in [ROCCurve, PRCurve, ConfusionMatrix]
}

MAX_ANALYSIS_SAMPLE_SIZE = 100000000


def split_and_preprocess(df: pd.DataFrame,
                         y_label: str,
                         x_labels: List[str],
                         df_synth: Optional[pd.DataFrame] = None,
                         sample_size: Optional[int] = None):
    """Split and preprocess the given dfs
       If synthetic data is provided, then training data should come from synthetic df
       and test from original df"""
    # If sample_size is not passed as the parameter of the function,
    # then set it to the minimum of MAX_ANALYSIS_SAMPLE_SIZE and data size
    if sample_size is None:
        sample_size = min(MAX_ANALYSIS_SAMPLE_SIZE, len(df))

    preprocessor = ModellingPreprocessor(target=y_label)
    if df_synth is not None:
        # Fit original and synthetic data together as preprocessor needs
        # to know all categorical variables
        preprocessor.fit(pd.concat((df, df_synth)))
        df_train_pre = preprocessor.transform(df_synth)  # Use all of the synthetic data as the train data
        df_test_pre = preprocessor.transform(df)  # Use all of the original data as the test data
    else:
        preprocessor.fit(df)
        df_train, df_test = sample_split_data(df,
                                              response_variable=y_label,
                                              explanatory_variables=x_labels,
                                              sample_size=sample_size)
        df_train_pre = preprocessor.transform(df_train)
        df_test_pre = preprocessor.transform(df_test)

    x_labels_pre = list(filter(lambda v: v != y_label, df_train_pre.columns))
    x_train = df_train_pre[x_labels_pre].to_numpy()
    y_train = df_train_pre[y_label].to_numpy()
    x_test = df_test_pre[x_labels_pre].to_numpy()
    y_test = df_test_pre[y_label].to_numpy()
    return x_train, y_train, x_test, y_test


def get_modelling_score(is_regression_task: bool,
                        x_train: np.ndarray,
                        y_train: np.ndarray,
                        x_test: np.ndarray,
                        y_test: np.ndarray,
                        estimator: Union[str, ClassifierMixin, RegressorMixin],
                        n_classes: Optional[int] = None):
    """Depending on the task (classification or regression), the score is evaluated."""
    if is_regression_task:
        metric = 'r2_score'
        task = 'regression'
        scores = regressor_scores(x_train, y_train, x_test, y_test,
                                  rgr=estimator, metrics=metric)
        score = scores[metric]
    else:
        metric = 'roc_auc' if n_classes == 2 else 'macro roc_auc'
        task = 'binary' if n_classes == 2 else f'multinomial [{n_classes}]'
        scores = classifier_scores(x_train, y_train, x_test, y_test,
                                   clf=estimator, metrics='roc_auc')
        score = scores['roc_auc']

    return score, metric, task


def predictive_modelling_score(
    df: pd.DataFrame,
    y_label: str,
    x_labels: Optional[List[str]],
    model: Union[str, BaseEstimator],
    df_synth: Optional[pd.DataFrame] = None,
    copy_model: bool = True,
    sample_size: Optional[int] = None,
    check: Check = ColumnCheck()
):
    """Calculates the R-squared or ROC AUC score of a given model trained
    on a given dataset. This function will fit a regressor or classifier
    depending on the datatype of the y_label. All necessary preprocessing
    (e.g standard scaling, one-hot encoding) is done in the function.
    The input data is automatically split into a training and testing set
    in order to evaluate the model performance.
    Args:
        data (pd.DataFrame): Input dataset.
        y_label (str): Name of the target variable column/response variable
                        to predict.
        x_labels (List[str], optional): Input column names/explanatory
                        variables. Defaults to None, in which case all
                        columns in the dataset except y_label will be used
                        as predictors.
        model (Union[str, sklearn.base.BaseEstimator]): One of 'Linear',
                        'GradientBoosting', 'RandomForrest', 'MLP',
                        'LinearSVM', or 'Logistic'. Note that 'Logistic'
                        only applies to categorical response variables.
                        Alternatively, a custom model class that inherits
                        from sklearn.base.BaseEstimator can be specified.
        synth_data (pd.DataFrame): Train the model on this synthetic data
                        but evaluate it's performance on the original.
    Returns:
        The score, metric ('r2' or 'roc_auc'), and the task ('regression',
        'binary', or 'multinomial')
    """

    score, metric, task = None, None, None

    # Remove rows with response variable as NaNs
    df = df[df[y_label].notna()].reset_index(drop=True)

    if df_synth is not None:
        df_synth = df_synth[df_synth[y_label].notna()].reset_index(drop=True)

    if y_label not in df.columns:
        raise ValueError('Response variable not in DataFrame.')

    if x_labels is None:
        x_labels = list(filter(lambda c: c != y_label, df.columns))
    else:
        available_columns = list(set(df.columns).intersection(set(x_labels + [y_label])))
        if not len([pred for pred in x_labels if pred in available_columns]):
            raise ValueError('Response/Predictor variables not in DataFrame.')
        df = df[available_columns]
        if df_synth is not None:
            df_synth = df_synth[available_columns]

    # Check target type
    is_regression_task = True
    n_classes = None
    if check.categorical(df[y_label]) is True:
        estimator = check_model_type(model, copy_model, 'clf')
        is_regression_task = False
        n_classes = len(df[y_label].dropna().unique())
    elif check.continuous(df[y_label]) is True:
        estimator = check_model_type(model, copy_model, 'rgr')
    else:
        raise ValueError(f"Can't understand y_label '{y_label}' type.")

    x_train, y_train, x_test, y_test = split_and_preprocess(df=df,
                                                            y_label=y_label,
                                                            x_labels=x_labels,
                                                            df_synth=df_synth,
                                                            sample_size=sample_size)
    estimator.fit(x_train, y_train)
    score, metric, task = get_modelling_score(is_regression_task=is_regression_task,
                                              x_train=x_train,
                                              y_train=y_train,
                                              x_test=x_test,
                                              y_test=y_test,
                                              estimator=estimator,
                                              n_classes=n_classes)

    return score, metric, task


class PredictiveModellingScore(DataFrameMetric):
    name = "predictive_modelling_score"

    def __init__(self,
                 model: str = None,
                 y_label: str = None,
                 x_labels: List[str] = None,
                 sample_size: Optional[int] = None,
                 df_test: Optional[pd.DataFrame] = None):
        self.model = model or 'Linear'
        self.y_label = y_label
        self.x_labels = x_labels
        self.sample_size = sample_size
        self.df_test = df_test

    def __call__(self,
                 df: pd.DataFrame):
        if len(df.columns) < 2:
            raise ValueError

        synth_score = None
        y_label = self.y_label or df.columns[-1]
        if self.x_labels is not None:
            x_labels = self.x_labels
        else:
            x_labels = [col for col in df.columns if col != y_label]

        # Train and evaluate of the train and the test part respectively of the original df
        score, metric, task = predictive_modelling_score(df=df,
                                                         y_label=y_label,
                                                         x_labels=x_labels,
                                                         model=self.model,
                                                         sample_size=self.sample_size)

        # If synthetic df is provided then get the synth_score by training on the
        # synthetic data and evaluating on the original data
        if self.df_test is not None:
            synth_score, _, _ = predictive_modelling_score(df=df,
                                                           y_label=y_label,
                                                           x_labels=x_labels,
                                                           model=self.model,
                                                           df_synth=self.df_test,
                                                           sample_size=self.sample_size)

        return score, synth_score, metric, task


def classifier_scores_from_df(df_train: Optional[pd.DataFrame],
                              df_test: pd.DataFrame,
                              target: str,
                              clf: ClassifierMixin,
                              metrics: Optional[Union[str, List[str]]] = 'roc_auc',
                              return_predicted: bool = False) -> Dict[str, Any]:
    x_labels = list(filter(lambda v: v != target, df_test.columns))

    if df_train is not None:
        x_train = df_train[x_labels].to_numpy()
        y_train = df_train[target].to_numpy()
    else:
        x_train = y_train = None
    x_test = df_test[x_labels].to_numpy()
    y_test = df_test[target].to_numpy()

    return classifier_scores(x_train, y_train, x_test, y_test, clf,
                             metrics, return_predicted)


def regressor_scores_from_df(df_train: Optional[pd.DataFrame],
                             df_test: pd.DataFrame,
                             target: str,
                             rgr: RegressorMixin, metrics: Optional[Union[str, List[str]]] = 'r2_score',
                             return_predicted: bool = False) -> Dict[str, Any]:
    x_labels = list(filter(lambda v: v != target, df_test.columns))

    if df_train is not None:
        x_train = df_train[x_labels].to_numpy()
        y_train = df_train[target].to_numpy()
    else:
        x_train = y_train = None
    x_test = df_test[x_labels].to_numpy()
    y_test = df_test[target].to_numpy()

    return regressor_scores(x_train, y_train, x_test, y_test, rgr, metrics,
                            return_predicted)


def _predict_classification_task(x_test: np.ndarray,
                                 y_test: np.ndarray,
                                 y_train: np.ndarray,
                                 clf: ClassifierMixin):
    multiclass: bool = False

    # Two classes classification
    if len(np.unique(y_train)) == 2:
        multiclass = False
        if hasattr(clf, 'predict_proba'):
            f_proba_test = clf.predict_proba(x_test)
            y_pred_test = np.argmax(f_proba_test, axis=1)
            f_proba_test = f_proba_test.T[1]
        else:
            y_pred_test = f_proba_test = clf.predict(x_test)

    # Multi-class classification
    else:
        multiclass = True
        oh = OneHotEncoder(sparse=False)
        oh.fit(np.concatenate((y_train, y_test)).reshape((-1, 1)))
        y_test_oh = oh.transform(y_test.reshape((-1, 1)))

        if hasattr(clf, 'predict_proba'):
            missing = [v for v in oh.categories_[0] if v not in clf.classes_]
            f_proba_test = clf.predict_proba(x_test)
            missing_idx = sorted([list(oh.categories_[0]).index(v)
                                 for v in missing], reverse=True)

            for idx in missing_idx:
                f_proba_test = np.insert(f_proba_test, idx, 0, 1)

            y_pred_test = np.array(oh.categories_[0])[np.argmax(f_proba_test,
                                                                axis=1)]
        else:
            y_pred_test = clf.predict(x_test)
            f_proba_test = oh.transform(y_pred_test.reshape(-1, 1))

        y_test_oh, f_proba_test = _remove_zero_column(y_test_oh, f_proba_test)

        # Normalise the prediction probabilities so that the probabilities
        # across classes add to 1.
        f_proba_test = _normalise_prob(f_proba_test)

    return multiclass, y_pred_test, f_proba_test


def classifier_scores(x_train: np.ndarray,
                      y_train: np.ndarray,
                      x_test: np.ndarray,
                      y_test: np.ndarray,
                      clf: ClassifierMixin,
                      metrics: Optional[Union[str, List[str]]] = 'roc_auc',
                      return_predicted: bool = False) -> Dict[str, Any]:

    all_classification_metrics: Dict[str, Union[Type[ClassificationMetric],
                                     Type[ClassificationPlotMetric]]]\
        = dict(CLASSIFICATION_METRICS, **CLASSIFICATION_PLOT_METRICS)

    if isinstance(metrics, str):
        metrics_dict = {
            metrics: all_classification_metrics[metrics]
        }
    elif isinstance(metrics, list):
        metrics_dict = {metric_name: all_classification_metrics[metric_name]
                        for metric_name in metrics}
    elif metrics is None:
        metrics_dict = {metric_name: CLASSIFICATION_METRICS[metric_name]
                        for metric_name in CLASSIFICATION_METRICS.keys()}
    else:
        raise TypeError("'metrics' type not recognized")

    # Check all metrics are supported
    missing_metrics = list(filter(lambda m: m not in
                                  all_classification_metrics.keys(),
                                  metrics_dict.keys()))
    if len(missing_metrics) > 0:
        raise ValueError("Can't compute following metrics: '{}'"
                         .format("', '".join(missing_metrics)))

    # Single class present in target
    if len(np.unique(y_train)) == 1:
        return {metric_name: 1. for metric_name in metrics_dict.keys()}

    # Check if fitted
    try:
        check_is_fitted(clf)
    except NotFittedError:
        clf.fit(x_train, y_train)

    multiclass, y_pred_test, f_proba_test = _predict_classification_task(x_test=x_test,
                                                                         y_test=y_test,
                                                                         y_train=y_train,
                                                                         clf=clf)
    results: Dict[str, Any] = dict()
    for metric_name, metric in metrics_dict.items():
        # metric() and metric.__call__() are the same,
        # but the latter raises lint error
        m_inst = metric(multiclass=multiclass)
        results[metric_name] = m_inst(y_true=y_test, y_pred=y_pred_test,
                                      y_pred_proba=f_proba_test)

    if return_predicted:
        results['predicted_values'] = y_pred_test

    return results


def regressor_scores(x_train: Optional[np.ndarray],
                     y_train: Optional[np.ndarray],
                     x_test: np.ndarray,
                     y_test: np.ndarray,
                     rgr: RegressorMixin,
                     metrics: Optional[Union[str, List[str]]] = 'r2_score',
                     return_predicted: bool = False) -> Dict[str, float]:

    if isinstance(metrics, str):
        metrics_dict: Dict[str, Type[RegressionMetric]]\
            = {metrics: REGRESSION_METRICS[metrics]}
    elif isinstance(metrics, list):
        if len(metrics) == 0:
            raise ValueError("Given empty array of metrics")
        metrics_dict = {metric_name: REGRESSION_METRICS[metric_name]
                        for metric_name in metrics}
    elif metrics is None:
        metrics_dict = {metric_name: REGRESSION_METRICS[metric_name]
                        for metric_name in REGRESSION_METRICS.keys()}
    else:
        raise TypeError("'metrics' type not recognized")

    # Check all metrics are supported
    missing_metrics = list(filter(lambda m: m not in REGRESSION_METRICS.keys(),
                                  metrics_dict.keys()))
    if len(missing_metrics) > 0:
        raise ValueError("Can't compute following metrics: '{}'"
                         .format("', '".join(missing_metrics)))

    try:
        check_is_fitted(rgr)
    except NotFittedError:
        if x_train is None or y_train is None:
            raise ValueError("'x_train' and 'y_train' must be given\
                              if the classifier has not been trained")
        rgr.fit(x_train, y_train)

    f_test = rgr.predict(x_test)

    results: Dict[str, Any] = dict()
    for metric_name, metric in metrics_dict.items():
        results[metric_name] = metric()(y_true=y_test, y_pred=f_test)

    if return_predicted:
        results['predicted_values'] = f_test

    return results


def _remove_zero_column(y1, y2):
    """Given two one-hot encodings, delete all columns from both arrays
    if they are all zeros for any of the two arrays.

    Args:
        y1, y2: Input arrays

    Returns:
        y1, y2: Output arrays

    """
    if len(y1.shape) != 2 or len(y2.shape) != 2:
        return y1, y2
    if y1.shape[1] != y2.shape[1]:
        raise ValueError(f"Mismatch of array shapes,\
                           {y1.shape} not compatible with {y2.shape}.")

    delete_index = np.where((y1 == 0).all(axis=0))

    y1 = np.delete(y1, delete_index, axis=1)
    y2 = np.delete(y2, delete_index, axis=1)
    return y1, y2


def _normalise_prob(f_proba):
    """Given probabilities across classes, normalise them
    so that they add to 1 for each row.

    Args:
        f_proba: Input probability array

    Returns:
        f_proba_normalised: Output normalised probability array

    """
    f_proba_normalised = normalize(f_proba, axis=1, norm='l1')
    idxs = np.where(~f_proba_normalised.any(axis=1))[0]
    num_classes = f_proba_normalised.shape[1]
    f_proba_normalised[idxs] = 1 / num_classes

    return f_proba_normalised




import pandas as pd
import pytest
from sklearn.ensemble import GradientBoostingRegressor, RandomForestClassifier

from synthesized_insight.modelling import ModellingPreprocessor
from synthesized_insight.modelling.utility import check_model_type, sample_split_data


@pytest.fixture(scope='module')
def df():
    df = pd.read_csv('tests/datasets/mini_credit_with_categoricals.csv')
    return df


def test_sample_split_data(df):
    response_variable = 'NumberOfDependents'
    df_train, df_test = sample_split_data(df, response_variable=response_variable)
    assert df_train is not None and df_test is not None
    train_unique_targets = df_train[response_variable].unique()
    test_unique_targets = df_test[response_variable].unique()
    assert all([test_target in train_unique_targets for test_target in test_unique_targets])


def test_modelling_preprocessor(df):
    data = pd.read_csv('tests/datasets/mini_credit_with_categoricals.csv')
    prep = ModellingPreprocessor(target='SeriousDlqin2yrs')
    assert prep.fit_transform(data) is not None


def test_check_model_type():
    assert isinstance(check_model_type('RandomForest', False, 'clf'), RandomForestClassifier)

    with pytest.raises(KeyError):
        check_model_type('AdaBoost', False, 'clf')

    with pytest.raises(ValueError):
        check_model_type(None, False, 'clf')

    assert isinstance(check_model_type('GradientBoosting', False, 'rgr'), GradientBoostingRegressor)

    with pytest.raises(KeyError):
        check_model_type('AdaBoost', False, 'rgr')

    with pytest.raises(ValueError):
        check_model_type(None, False, 'rgr')



import numpy as np
import pandas as pd
import pytest

from src.synthesized_insight.check import Check, ColumnCheck
from src.synthesized_insight.metrics import (
    BinomialDistanceTest,
    EarthMoversDistance,
    HellingerDistance,
    KendallTauCorrelationTest,
    KolmogorovSmirnovDistanceTest,
    KruskalWallisTest,
    SpearmanRhoCorrelationTest,
)
from src.synthesized_insight.metrics.base import TwoColumnMetric
from src.synthesized_insight.metrics.statistical_tests import BootstrapTest, PermutationTest


@pytest.fixture(scope='module')
def df():
    df = pd.read_csv("tests/datasets/mini_compas.csv")
    return df


@pytest.fixture(scope='module')
def group1(df):
    pred1 = df["Ethnicity"] == "Caucasian"
    target_attr = "RawScore"
    group1 = df[pred1][target_attr]
    return group1


class Distance(TwoColumnMetric):
    @classmethod
    def check_column_types(cls, sr_a: pd.Series, sr_b: pd.Series, check: Check = ColumnCheck()):
        return True

    def _compute_metric(self, sr_a, sr_b):
        metric_val = np.sum((sr_a - sr_b)**2)
        return metric_val


@pytest.fixture
def data1():
    return np.random.normal(0, 1, 1000)


@pytest.fixture
def data2():
    return np.random.normal(1, 1, 1000)


def test_distance(data1, data2):
    assert Distance()(pd.Series(data1), pd.Series(data2)) > 0


def test_pvalue(data1, data2):
    bootstrap_test = BootstrapTest(metric_cls_obj=Distance())
    _, p_value = bootstrap_test(pd.Series(data1), pd.Series(data2))
    assert p_value >= 0
    assert p_value <= 1


def test_binomial_distance_metric():
    metric = BinomialDistanceTest()
    data = (pd.Series([1, 0, 0]), pd.Series([1, 0, 1]))
    metric_value, p_value = metric(data[0], data[0])
    assert metric_value == 0
    assert p_value == 1

    metric_value, p_value = metric(data[0], data[1])
    assert np.abs(metric_value) > 0
    assert p_value <= 1
    assert p_value >= 0


def test_binomial_p_value():
    assert BinomialDistanceTest()(pd.Series([1, 1]), pd.Series([0, 0]))[1] == 0
    assert BinomialDistanceTest()(pd.Series([1, 0]), pd.Series([1, 0]))[1] == 1
    assert BinomialDistanceTest()(pd.Series([1, 0, 1, 1]), pd.Series([1, 0, 1, 0]))[1] == 0.625


def test_ks_distance():
    sr_a = pd.Series(np.random.normal(0, 1, 100), name='a')
    sr_b = pd.Series(['a', 'b', 'c', 'd'], name='b')

    ksd_test = KolmogorovSmirnovDistanceTest()
    assert ksd_test(sr_a, sr_a)[0] is not None
    assert ksd_test(sr_b, sr_b)[0] is None


def test_kt_correlation():
    np.random.seed(42)
    sr_a = pd.Series(np.random.normal(0, 1, 100), name='a')
    sr_b = pd.Series(np.random.normal(0, 1, 5), name='b')
    sr_c = pd.Series(sr_b.values + np.random.normal(0, 0.8, 5), name='c')
    sr_d = pd.Series(['a', 'b', 'c', 'd'], name='d')

    kt_corr_test = KendallTauCorrelationTest(max_p_value=0.05)

    assert kt_corr_test(sr_a, sr_a)[0] is not None
    assert kt_corr_test(sr_b, sr_c)[0] is None
    assert kt_corr_test(sr_c, sr_d)[0] is None


def test_kt_correlation_string_numbers():
    sr_a = pd.Series([0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14], name="a")
    sr_b = pd.Series(['1.4', '2.1', '', '4.1', '3.9', '4.4', '5.1', '6.0', '7.5', '9', '11.4', '12.1', '', '14.1', '13.9'], name="b")
    kendell_tau_correlation_test = KendallTauCorrelationTest(max_p_value=0.05)

    # df['TotalCharges'].dtype is Object in this case, eg. "103.4" instead of 103.4
    kt1 = kendell_tau_correlation_test(sr_a=sr_a, sr_b=sr_b)

    sr_b = pd.to_numeric(sr_b, errors='coerce')
    kt2 = kendell_tau_correlation_test(sr_a=sr_a, sr_b=sr_b)
    assert abs(kt1[0] - kt2[0]) < 0.01


def test_kruskal_wallis(group1):
    kruskal_wallis_test = KruskalWallisTest()
    assert kruskal_wallis_test(pd.Series(np.arange(10)), pd.Series(np.arange(10)))[0] == 0
    assert abs(kruskal_wallis_test(group1, group1)[0]) < 1e-6


def test_spearmans_rho():
    sr_a = pd.Series(np.random.normal(0, 1, 100), name='a')
    sr_b = pd.Series([1, 2, 3, 4, 5], name='b')
    sr_c = pd.Series([3, 5, 1, 3, 2], name='c')
    sr_d = pd.Series(['a', 'b', 'c', 'd'], name='d')

    spearman_rho_test = SpearmanRhoCorrelationTest(max_p_value=0.05)
    assert spearman_rho_test(sr_a, sr_a)[0] is not None
    assert spearman_rho_test(sr_b, sr_c)[0] is None
    assert spearman_rho_test(sr_c, sr_d)[0] is None


@pytest.mark.parametrize(
    'metric, data, alternative', [
        (EarthMoversDistance(), (pd.Series(['a', 'b', 'c']), pd.Series(['c', 'b', 'a'])), 'greater'),
        (HellingerDistance(), (pd.Series([1, 2, 3]), pd.Series([0, 0, 0])), 'less')
    ])
def test_bootstrap(metric, data, alternative):
    bootstrap = BootstrapTest(metric_cls_obj=metric)
    metric_val, p_value = bootstrap(data[0], data[0])
    assert metric_val == 0
    assert p_value == 1

    bootstrap = BootstrapTest(metric_cls_obj=metric, alternative=alternative)
    metric_val, p_value = bootstrap(data[1], data[1])
    assert metric_val == 0
    assert p_value == 1 if alternative != 'less' else p_value == 0

    _, p_value = bootstrap(data[0], data[1])
    assert p_value <= 1
    assert p_value >= 0


@pytest.mark.parametrize(
    'metric, data, alternative', [
        (EarthMoversDistance(), (pd.Series(['a', 'b', 'c']), pd.Series(['c', 'b', 'a'])), 'greater'),
        (HellingerDistance(), (pd.Series([1, 2, 3]), pd.Series([0, 0, 0])), 'less')
    ])
def test_permutation(metric, data, alternative):
    perm_test = PermutationTest(metric_cls_obj=metric)
    metric_val, p_value = perm_test(data[0], data[0])
    assert metric_val == 0
    assert p_value == 1

    perm_test = PermutationTest(metric_cls_obj=metric, alternative=alternative)
    metric_val, p_value = perm_test(data[1], data[1])
    assert metric_val == 0
    assert p_value == 1 if alternative != 'less' else p_value == 0

    _, p_value = perm_test(data[0], data[1])
    assert p_value <= 1
    assert p_value >= 0



from itertools import combinations

import numpy as np
import pandas as pd
import pytest

from synthesized_insight.check import ColumnCheck
from synthesized_insight.metrics import (
    CramersV,
    EarthMoversDistance,
    EarthMoversDistanceBinned,
    HellingerDistance,
    JensenShannonDivergence,
    KendallTauCorrelationTest,
    KullbackLeiblerDivergence,
    Mean,
    Norm,
    R2Mcfadden,
    StandardDeviation,
)

mean = Mean()
std_dev = StandardDeviation()
cramers_v = CramersV()
emd = EarthMoversDistance()
kendell_tau_correlation_test = KendallTauCorrelationTest()
hellinger_distance = HellingerDistance()
kl_divergence = KullbackLeiblerDivergence()
js_divergence = JensenShannonDivergence()
r2_mcfadden = R2Mcfadden()
norm = Norm()
norm_ord1 = Norm(ord=1)


@pytest.fixture(scope='module')
def df():
    df = pd.read_csv("tests/datasets/mini_compas.csv")
    return df


@pytest.fixture(scope='module')
def group1(df):
    pred1 = df["Ethnicity"] == "Caucasian"
    target_attr = "RawScore"
    group1 = df[pred1][target_attr]
    return group1


@pytest.fixture(scope='module')
def group2(df):
    pred2 = df["Ethnicity"] == "African-American"
    target_attr = "RawScore"
    group2 = df[pred2][target_attr]
    return group2


@pytest.fixture(scope='module')
def group3(group2):
    group3 = group2.sort_values()[len(group2) // 2:]
    return group3


@pytest.fixture
def data1():
    return np.random.normal(0, 1, 1000)


@pytest.fixture
def data2():
    return np.random.normal(1, 1, 1000)


def test_mean():
    sr_a = pd.Series(np.arange(100), name='a')
    val_a = mean(sr=sr_a)
    assert val_a == 49.5

    sr_b = pd.Series(np.datetime64('2020-01-01') + np.arange(0, 3, step=1).astype('m8[D]'), name='b')
    val_b = mean(sr=sr_b)
    assert val_b == np.datetime64('2020-01-02')

    sr_c = pd.Series(['a', 'b', 'c', 'd'], name='c')
    val_c = mean(sr=sr_c)
    assert val_c is None


def test_standard_deviation():
    sr_a = pd.Series(np.random.normal(0, 1, 100), name='a')
    val_a = std_dev(sr=sr_a)
    assert val_a is not None

    sr_b = pd.Series(np.datetime64('2020-01-01') + np.arange(0, 20, step=1).astype('m8[D]'), name='b')
    val_b = std_dev(sr=sr_b)
    assert val_b is not None

    sr_c = pd.Series(['a', 'b', 'c', 'd'], name='c')
    val_c = std_dev(sr=sr_c)
    assert val_c is None


def test_em_distance():
    sr_a = pd.Series(np.random.normal(0, 1, 100), name='a')
    sr_b = pd.Series(['a', 'b', 'c', 'd'], name='b')

    assert emd(sr_a, sr_a) is None
    assert emd(sr_b, sr_b) is not None


def test_cramers_v_basic():
    sr_a = pd.Series([1, 2, 3, 1, 2, 3, 1, 2, 3] * 100, name='a')
    sr_b = pd.Series([1, 2, 3, 2, 3, 1, 3, 1, 2] * 100, name='b')

    assert cramers_v(sr_a, sr_a) > 0.99  # This metric -> cramers_v for large N (makes it more robust to outliers)
    assert cramers_v(sr_a, sr_b) == 0.

    sr_c = pd.Series(np.random.normal(0, 1, 1000), name='c')
    assert cramers_v(sr_c, sr_c) is None


def test_cramers_v_compas(df):
    check = ColumnCheck()
    continuous_cols, categorical_cols = [], []

    for col in df.columns:
        if check.continuous(df[col]):
            continuous_cols.append(col)
        elif check.categorical(df[col]):
            categorical_cols.append(col)

    for col_grp in combinations(categorical_cols, 2):
        assert(cramers_v(df[col_grp[0]], df[col_grp[1]]) is not None)


def test_repr():
    metric = EarthMoversDistance()
    assert repr(metric) == 'EarthMoversDistance()'


def test_str():
    metric = EarthMoversDistance()
    metric.name = 'earth_movers_distance'
    assert str(metric) == 'earth_movers_distance'


def test_kl_divergence(group1):
    assert kl_divergence(pd.Series([1, 0]), pd.Series([1, 0])) == 0
    assert kl_divergence(pd.Series([1, 1]), pd.Series([0, 0])) == float("inf")

    assert kl_divergence(group1, group1) == 0


def test_js_divergence(group1, group2, group3):
    assert js_divergence(pd.Series([1, 0]), pd.Series([1, 0])) == 0

    assert js_divergence(group1, group1) == 0
    assert js_divergence(group1, group3) > js_divergence(group1, group2)


def test_norm(group1, group2, group3):
    assert norm(pd.Series([1, 0]), pd.Series([1, 0])) == 0
    assert norm_ord1(pd.Series([1]), pd.Series([0])) == 2

    assert norm(group1, group1) == 0
    assert norm(group1, group3) > norm(group1, group2)


def test_hellinger(group1, group2, group3):
    assert hellinger_distance(pd.Series([1, 0]), pd.Series([1, 0])) == 0
    assert hellinger_distance(pd.Series([1]), pd.Series([0])) == 1

    assert hellinger_distance(group1, group1) == 0
    assert hellinger_distance(group1, group3) > hellinger_distance(group1, group2)


def test_r2_mcfadden_correlation():
    sr_a = pd.Series(np.random.normal(0, 1, 100), name='x')
    sr_a_dates = pd.Series(pd.date_range('01/01/01', '01/12/01', name='x'))
    sr_b = pd.Series(np.random.choice([1, 0], 100), name='y')

    assert r2_mcfadden(sr_a, sr_a) is None  # continuous, continuous -> None

    assert r2_mcfadden(sr_b, sr_b) is None  # categorical, categorical -> None

    assert r2_mcfadden(sr_a_dates, sr_b) is None  # continuous date, categorical -> None

    assert r2_mcfadden(sr_b, sr_a) is not None  # categorical, continuous -> not None

    credit_df = pd.read_csv('tests/datasets/mini_credit.csv').sample(500)
    assert r2_mcfadden(sr_a=credit_df['SeriousDlqin2yrs'], sr_b=credit_df['age']) is not None
    assert r2_mcfadden(sr_a=credit_df['MonthlyIncome'], sr_b=credit_df['DebtRatio']) is None
    assert r2_mcfadden(sr_a=credit_df['SeriousDlqin2yrs'], sr_b=credit_df['MonthlyIncome']) is not None


def test_emd_distance_binned():

    def compare_and_log(x, y, bin_edges, val):
        emdb = EarthMoversDistanceBinned(bin_edges=bin_edges)
        metric_val = emdb(x, y)
        assert np.isclose(metric_val, val, rtol=0.1)

    compare_and_log(pd.Series([1, 2, 3]), pd.Series([1, 0, 3]), bin_edges=[0, 1, 2, 3], val=0.333)

    a = pd.Series(np.random.normal(loc=10, scale=1.0, size=10000))
    b = pd.Series(np.random.normal(loc=14, scale=1.0, size=10000))

    bin_edges = np.histogram_bin_edges(np.concatenate((a, b), axis=0), bins=100)
    x, _ = np.histogram(a, bins=bin_edges)
    y, _ = np.histogram(b, bins=bin_edges)
    compare_and_log(pd.Series(x), pd.Series(y), bin_edges, 4.0)
    compare_and_log(pd.Series([0, 3, 6, 14, 3]), pd.Series([1, 0, 8, 21, 1]), None, 0.20)
    compare_and_log(pd.Series([0, 3, 6, 14, 3]), pd.Series([0, 3, 6, 14, 3]), None, 0.0)
    compare_and_log(pd.Series([0, 0, 0, 0]), pd.Series([0, 3, 6, 14]), None, 1.0)
    compare_and_log(pd.Series([0, 0, 0, 0]), pd.Series([0, 0, 0, 0]), None, 0.0)



from typing import Callable, Tuple, Union

import numpy as np
import pandas as pd
from scipy.stats import binom_test, kendalltau, kruskal, ks_2samp, spearmanr

from ..check import Check, ColumnCheck
from .base import TwoColumnMetric, TwoColumnTest
from .utils import bootstrap_binned_statistic, bootstrap_statistic


class BinomialDistanceTest(TwoColumnTest):
    """Binomial distance test between two binary variables.

    The metric value ranges from 0 to 1, where a value of 0 indicates there is no association between the variables,
    and 1 indicates maximal association (i.e one variable is completely determined by the other).
    """
    name = "binomial_distance"

    @classmethod
    def check_column_types(cls, sr_a: pd.Series, sr_b: pd.Series, check: Check = ColumnCheck()) -> bool:
        return sr_a.append(sr_b).nunique() == 2

    def _compute_test(self, sr_a: pd.Series, sr_b: pd.Series) -> Tuple[Union[int, float, None], Union[int, float, None]]:
        """Calculate binomial metric and exact p-value for an observed binomial proportion of a sample."""

        metric_value = sr_a.mean() - sr_b.mean()

        p_obs = sr_a.mean()
        p_null = sr_b.mean()
        n = len(sr_a)

        k = np.ceil(p_obs * n)
        p_value = binom_test(k, n, p_null, self.alternative)
        return metric_value, p_value


class KolmogorovSmirnovDistanceTest(TwoColumnTest):
    """Kolmogorov-Smirnov distance test between two continuous variables.
    The statistic ranges from 0 to 1, where a value of 0 indicates the two variables follow identical distributions,
    and a value of 1 indicates they follow completely different distributions.

    """
    name = "kolmogorov_smirnov_distance"

    @classmethod
    def check_column_types(cls, sr_a: pd.Series, sr_b: pd.Series, check: Check = ColumnCheck()) -> bool:
        if not check.continuous(sr_a) or not check.continuous(sr_b):
            return False
        return True

    def _compute_test(self, sr_a: pd.Series, sr_b: pd.Series) -> Tuple[Union[int, float, None], Union[int, float, None]]:
        """Calculate the metric and p-value.
        Args:
            sr_a (pd.Series): values of a continuous variable.
            sr_b (pd.Series): values of another continuous variable to compare.
        Returns:
            The Kolmogorov-Smirnov distance between sr_a and sr_b,
            and p-value on the test statistics
        """
        column_old_clean = pd.to_numeric(sr_a, errors='coerce').dropna()
        column_new_clean = pd.to_numeric(sr_b, errors='coerce').dropna()
        if len(column_old_clean) == 0 or len(column_new_clean) == 0:
            return np.nan, np.nan
        ks_distance, p_value = ks_2samp(column_old_clean, column_new_clean, alternative=self.alternative)
        return ks_distance, p_value


class KruskalWallisTest(TwoColumnTest):
    """Kruskal Wallis distance test between two numerical variables.

    The statistic ranges from 0 to 1, where a value of 0 indicates there is no association between the variables,
    and 1 indicates maximal association (i.e one variable is completely determined by the other).
    """
    name = "kruskal_wallis"

    @classmethod
    def check_column_types(cls, sr_a: pd.Series, sr_b: pd.Series, check: Check = ColumnCheck()) -> bool:
        if not check.continuous(sr_a) or not check.continuous(sr_b):
            return False
        return True

    def _compute_test(self, sr_a: pd.Series, sr_b: pd.Series) -> Tuple[Union[int, float, None], Union[int, float, None]]:
        """Calculate the metric and the p_value.
        Args:
            sr_a (pd.Series): values of a numerical variable.
            sr_b (pd.Series): values of numerical variable.
        Returns:
            The Kruskal Wallis distance between sr_a and sr_b,
            and p-value on the test statistics
        """
        metric_value, p_value = kruskal(sr_a, sr_b)
        return metric_value, p_value


class KendallTauCorrelationTest(TwoColumnTest):
    """Kendall's Tau correlation coefficient test between ordinal variables.
    The statistic ranges from -1 to 1, indicating the strength and direction of the relationship
    between the two variables.

    Args:
        max_p_value (float, optional): Returns None if p-value from test is above this threshold.
    """
    name = "kendall_tau_correlation"
    symmetric = True

    def __init__(self,
                 check: Check = ColumnCheck(),
                 alternative: str = 'two-sided',
                 max_p_value: float = 1.0):
        super().__init__(check, alternative)
        self.max_p_value = max_p_value

    @classmethod
    def check_column_types(cls, sr_a: pd.Series, sr_b: pd.Series, check: Check = ColumnCheck()) -> bool:
        # Given columns should be both categorical or both ordinal
        if (check.ordinal(sr_a) and check.ordinal(sr_b)):
            return True
        return False

    def _compute_test(self, sr_a: pd.Series, sr_b: pd.Series) -> Tuple[Union[int, float, None], Union[int, float, None]]:
        """Returns kendall-tau correlation and p-value"""

        sr_a = pd.to_numeric(sr_a, errors='coerce')
        sr_b = pd.to_numeric(sr_b, errors='coerce')
        corr, p_value = kendalltau(sr_a.values, sr_b.values, nan_policy='omit')
        if p_value is not None and p_value <= self.max_p_value:
            return corr, p_value
        else:
            return None, p_value


class SpearmanRhoCorrelationTest(TwoColumnTest):
    """Spearman's rank correlation coefficient test between ordinal variables.
    The statistic ranges from -1 to 1, measures the strength and direction of monotonic
    relationship between two ranked (ordinal) variables.

    Args:
        max_p_value (float, optional): Returns None if p-value from test is above this threshold.
    """
    name = "spearman_rho_correlation"

    def __init__(self,
                 check: Check = ColumnCheck(),
                 alternative: str = 'two-sided',
                 max_p_value: float = 1.0):
        super().__init__(check, alternative)
        self.max_p_value = max_p_value

    @classmethod
    def check_column_types(cls, sr_a: pd.Series, sr_b: pd.Series, check: Check = ColumnCheck()) -> bool:
        if not check.ordinal(sr_a) or not check.ordinal(sr_b):
            return False
        return True

    def _compute_test(self, sr_a: pd.Series, sr_b: pd.Series) -> Tuple[Union[int, float, None], Union[int, float, None]]:
        """Return spearman-rho correlation and p-value"""

        x = sr_a.values
        y = sr_b.values
        if self.check.infer_dtype(sr_a).dtype.kind == 'M':
            x = pd.to_numeric(pd.to_datetime(x, errors='coerce'), errors='coerce')
        if self.check.infer_dtype(sr_b).dtype.kind == 'M':
            y = pd.to_numeric(pd.to_datetime(y, errors='coerce'), errors='coerce')
        corr, p_value = spearmanr(x, y, nan_policy='omit')
        if p_value is not None and p_value <= self.max_p_value:
            return corr, p_value
        else:
            return None, p_value


class BootstrapTest:
    """Performs bootstrap test

    Args:
        metric_cls_obj: The instantiated metric object for which the test if performed
        binned: If binning is done
    """
    name = "bootstrap_test"

    def __init__(self,
                 metric_cls_obj: TwoColumnMetric,
                 alternative: str = 'two-sided',
                 binned: bool = False):
        self.alternative = alternative
        self.metric_cls_obj = metric_cls_obj
        self.binned = binned

    @classmethod
    def check_column_types(cls, sr_a: pd.Series, sr_b: pd.Series, check: Check = ColumnCheck()) -> bool:
        return True  # metric object provided during initialization will perform the check while computing the metrics

    def _bootstrap_pvalue(self,
                          t_obs: float,
                          t_distribution: np.ndarray) -> float:
        """
        Calculate a p-value using a bootstrapped test statistic distribution

        Args:
            t_obs: Observed value of the test statistic.
            t_distribution: Samples of test statistic distribution under the null hypothesis.

        Returns:
            The p-value under the null hypothesis.
        """

        n_samples = len(t_distribution)

        if self.alternative == 'two-sided':
            p = np.sum(np.abs(t_distribution) >= np.abs(t_obs)) / n_samples

        elif self.alternative == 'greater':
            p = np.sum(t_distribution >= t_obs) / n_samples

        else:
            p = np.sum(t_distribution < t_obs) / n_samples

        return p

    def __call__(self, sr_a: pd.Series, sr_b: pd.Series) -> Tuple[Union[int, float, None], Union[int, float, None]]:
        """Return metric_value and p-value for this metric using a bootstrap test.

        The null hypothesis is that both data samples are from the same distribution."""

        metric_value = self.metric_cls_obj(sr_a, sr_b)
        if metric_value is None:
            return None, None

        if not self.binned:
            ts_distribution = bootstrap_statistic((sr_a, sr_b),
                                                  lambda x, y: self.metric_cls_obj(x, y),
                                                  n_samples=1000)
        else:
            ts_distribution = bootstrap_binned_statistic((sr_a, sr_b),
                                                         lambda x, y: self.metric_cls_obj(x, y),
                                                         n_samples=1000)
        p_value = self._bootstrap_pvalue(metric_value, ts_distribution)
        return metric_value, p_value


class PermutationTest:
    """Performs permutation test

    Args:
        metric_cls_obj: The instantiated metric object for which the test if performed
        n_perms: No. of permutations to be done
    """
    name = "permutation_test"

    def __init__(self,
                 metric_cls_obj: TwoColumnMetric,
                 alternative: str = 'two-sided',
                 n_perms: int = 100):
        self.alternative = alternative
        self.metric_cls_obj = metric_cls_obj
        self.n_perms = n_perms

    def _permutation_test(self,
                          x: pd.Series,
                          y: pd.Series,
                          t: Callable[[pd.Series, pd.Series], float]) -> float:
        """
        Perform a two sample permutation test.
        Determines the probability of observing t(x, y) or greater under the null hypothesis that x
        and y are from the same distribution.
        Args:
            x: First data sample.
            y: Second data sample.
            t: Callable that returns the test statistic.
        Returns:
            The p-value of t_obs under the null hypothesis.
        """

        if t is None:
            raise ValueError("Callable function required, can't be None")

        t_obs = t(x, y)
        pooled_data = np.concatenate((x, y))
        t_null = np.empty(self.n_perms)

        for i in range(self.n_perms):
            perm = np.random.permutation(pooled_data)
            x_sample = perm[:len(x)]
            y_sample = perm[len(x):]
            t_null[i] = t(pd.Series(x_sample), pd.Series(y_sample))

        if self.alternative == 'two-sided':
            p = np.sum(np.abs(t_null) >= np.abs(t_obs)) / self.n_perms

        elif self.alternative == 'greater':
            p = np.sum(t_null >= t_obs) / self.n_perms

        else:
            p = np.sum(t_null < t_obs) / self.n_perms

        return p

    def __call__(self, sr_a: pd.Series, sr_b: pd.Series) -> Tuple[Union[int, float, None], Union[int, float, None]]:
        """Return metric_value and p-value for this metric using a permutation test.

        The null hypothesis is that both data samples are from the same distribution."""

        metric_value = self.metric_cls_obj(sr_a, sr_b)
        if metric_value is None:
            return None, None

        p_value = self._permutation_test(sr_a,
                                         sr_b,
                                         lambda x, y: self.metric_cls_obj(x, y))
        return metric_value, p_value



from typing import Any, List, Optional, Union

import numpy as np
import pandas as pd
from scipy.stats import norm
from sklearn.base import BaseEstimator

from ..check import Check, ColumnCheck
from ..modelling import check_model_type
from .base import TwoDataFrameMetric
from .modelling_metrics import split_and_preprocess
from .privacy_cap_scorer import DistanceCAPScorer, GeneralizedCAPScorer


class AttributeInferenceAttackML(TwoDataFrameMetric):
    """Computes the privacy score between two dataframes using ML model

    The privacy score is calculated by fitting a ML model on the synthetic dataset. The fitted
    model is then used to calculate the privacy score of the sensitive column in the original dataset.

    Score lies in the range of 0 to 1, 0 means negligible privacy and 1 means absolute privacy.

    Args:
        model: Str or Base estimator
        sensitive_col: Sensitive column to be predicted
        predictors: List of predictor column names. If not provided, then all columns
                    except the sensitive column is treated as predictors

    Returns:
        Privacy score between 0 and 1
    """
    def __init__(self,
                 model: Union[BaseEstimator, str],
                 sensitive_col: str,
                 predictors: Optional[List[str]] = None,
                 check: Check = ColumnCheck()) -> None:
        self.model = model
        self.sensitive_col = sensitive_col
        self.predictors = predictors
        self.check = check

    def _privacy_score_categorical(self, true: pd.Series, pred: pd.Series) -> float:
        """Computes privacy score as the ratio of total true and pred values that do not match

        Args:
            true: True values of the sensitive attribute
            pred: Predicted values of the sensitive attribute

        Returns:
            The privacy score between 0 and 1
            0 means negligible privacy and 1 means absolute privacy.
        """
        num_rows = len(true)
        matches = 0
        for idx in range(num_rows):
            if true[idx] == pred[idx]:
                matches += 1

        return 1 - (matches / num_rows)

    def _privacy_score_numerical(self, true: pd.Series, pred: pd.Series, lp: float = 0.5) -> float:
        """Computes privacy score by measuring the distance between each set of true and pred values of
        the sensitive column using CDF

        Firstly, the norm is fit on the true data. Then, the distance for each pair of true and pred record
        is the difference in the percentile of these values according to cdf raised by pth power. The distance is
        averaged over all the records to give a final score.

        Args:
            true: True values of the sensitive column in the original dataset
            pred: Predicted values of the sensitive column of the original dataset

        Returns:
            Privacy score between 0 and 1
        """
        num_rows = len(true)
        norm.fit(true)
        dist = 0
        for idx in range(num_rows):
            percentiles = norm.cdf(np.array([pred[idx], true[idx]]))
            dist += abs(percentiles[0] - percentiles[1])**lp

        return dist / num_rows

    def _calculate_privacy_score(self,
                                 is_discrete_target: bool,
                                 estimator: BaseEstimator,
                                 x_test: pd.DataFrame,
                                 y_test: pd.Series) -> float:
        """Calculate the privacy score using fitted ML model

        Args:
            is_discrete_target: If target is discrete or not
            estimator: ML model
            x_test: Preprocessed predictors in the original dataset
            y_test: Sensitive target in the original dataset

        Returns:
            Privacy score between 0 and 1
        """
        if is_discrete_target:
            if hasattr(estimator, 'predict_proba'):
                f_proba_test = estimator.predict_proba(x_test)
                y_pred_test = np.argmax(f_proba_test, axis=1)
            else:
                y_pred_test = estimator.predict(x_test)
            privacy_score = self._privacy_score_categorical(y_test, y_pred_test)
        else:
            y_pred_test = estimator.predict(x_test)
            privacy_score = self._privacy_score_numerical(y_test, y_pred_test)

        return privacy_score

    def __call__(self, orig_df: pd.DataFrame, synth_df: pd.DataFrame) -> float:
        """Computes the privacy score of the sensitive column of the original dataset based on the synthetic data

        Fits the ML model on the synthetic data. The trained model is used to predict the sensitive column
        of the original data. Finally, numerical or categorical privacy scorer is called depending on
        whether the sensitive column is a numerical or a categorical column.

        Args:
            orig_df: The original dataset
            synth_df: The synthetic dataset
            df_model: (Optional) The DataFrame model corresponding to synthetic dataset

        Returns:
            Privacy score between 0 and 1
        """
        if self.predictors is not None and not all(col in orig_df and col in synth_df for col in self.predictors):
            raise ValueError('All the predictors are not in original data and/or synthetic data')

        if not (self.sensitive_col in orig_df and self.sensitive_col in synth_df):
            raise ValueError('Sensitive column is not present in original data and/or synthetic data')

        if self.predictors is not None:
            orig_df = orig_df[self.predictors + [self.sensitive_col]]
            synth_df = synth_df[self.predictors + [self.sensitive_col]]
        else:
            self.predictors = synth_df.columns.intersection(orig_df.columns).tolist()

        estimator = None

        # Remove rows with response column value as NaN
        orig_df.dropna(subset=[self.sensitive_col], inplace=True)
        synth_df.dropna(subset=[self.sensitive_col], inplace=True)

        # Get the estimator
        is_discrete_target = True
        if self.check.categorical(orig_df[self.sensitive_col]) is True:
            estimator = check_model_type(model=self.model, copy_model=True, task='clf')
        elif self.check.continuous(orig_df[self.sensitive_col]) is True:
            is_discrete_target = False
            estimator = check_model_type(model=self.model, copy_model=True, task='rgr')
        else:
            raise ValueError(f"Can't understand y_label '{self.sensitive_col}' type.")

        x_train, y_train, x_test, y_test = split_and_preprocess(df=orig_df,
                                                                y_label=self.sensitive_col,
                                                                x_labels=self.predictors,
                                                                df_synth=synth_df)
        estimator.fit(x_train, y_train)
        privacy_score = self._calculate_privacy_score(is_discrete_target,
                                                      estimator, x_test, y_test)
        return privacy_score


class AttributeInferenceAttackCAP(TwoDataFrameMetric):
    """Computes the privacy score between two dataframes using CAP (Correct Attribution Probability) method

    Args:
        model: 'GeneralizedCAP' or 'DistanceCAP'
        sensitive_col: Name of the sensitive column
        predictors: List of predictor column names. If not provided, then all columns
                    except the sensitive column is treated as predictors

    Returns:
        Privacy score between 0 and 1
    """
    def __init__(self,
                 model: str,
                 sensitive_col: str,
                 predictors: Optional[List[str]] = None) -> None:
        self.model: Any = None
        if model == 'GeneralizedCAP':
            self.model = GeneralizedCAPScorer()
        elif model == 'DistanceCAP':
            self.model = DistanceCAPScorer()
        else:
            raise ValueError('Provided model string is not valid')
        self.predictors = predictors
        self.sensitive_col = sensitive_col

    def __call__(self, orig_df: pd.DataFrame, synth_df: pd.DataFrame) -> float:
        """Computes the privacy score of the sensitive column of the original dataset based on the synthetic data

        Fits the model on the synthetic data by mapping predictor key to sensitive attribute values of the synthetic dataset.
        Once fitted, the score method calculates the privacy of the sensitive attribute of the original dataset

        Args:
            orig_df: The original dataset
            synth_df: The synthetic dataset
            df_model: (Optional) The DataFrame model corresponding to synthetic dataset

        Return:
            Privacy score between 0 and 1
        """
        if self.predictors is not None and not all(col in orig_df and col in synth_df for col in self.predictors):
            raise ValueError('All the predictors are not in original data and/or synthetic data')

        if not (self.sensitive_col in orig_df and self.sensitive_col in synth_df):
            raise ValueError('Sensitive column is not present in original data and/or synthetic data')

        if self.predictors is None:
            predictors = [col for col in orig_df.columns if col != self.sensitive_col]
        else:
            predictors = self.predictors

        self.model.fit(synth_df[predictors], synth_df[self.sensitive_col])
        return self.model.score(orig_df[predictors], orig_df[self.sensitive_col])



from typing import Callable, Optional, Tuple, Union

import numpy as np
import pandas as pd

from ..check import Check, ColumnCheck


def zipped_hist(
    data: Tuple[pd.Series, ...],
    check: Check = ColumnCheck(),
    bin_edges: Optional[np.ndarray] = None,
    normalize: bool = True,
    ret_bins: bool = False,
) -> Union[Tuple[pd.Series, ...], Tuple[Tuple[pd.Series, ...], Optional[np.ndarray]]]:
    """Bins a tuple of series' and returns the aligned histograms.
    Args:
        data (Tuple[pd.Series, ...]):
            A tuple consisting of the series' to be binned. All series' must have the same dtype.
        bin_edges (Optional[np.ndarray], optional):
            Bin edges to bin continuous data by. Defaults to None.
        normalize (bool, optional):
            Normalize the histograms, turning them into pdfs. Defaults to True.
        ret_bins (bool, optional):
            Returns the bin edges used in the histogram. Defaults to False.
        distr_type (Optional[str]):
            The type of distribution of the target attribute. Can be "categorical" or "continuous".
            If None the type of distribution is inferred based on the data in the column.
            Defaults to None.
    Returns:
        Union[Tuple[np.ndarray, ...], Tuple[Tuple[np.ndarray, ...], Optional[np.ndarray]]]:
            A tuple of np.ndarrays consisting of each histogram for the input data.
            Additionally returns bins if ret_bins is True.
    """

    joint = pd.concat(data)
    is_continuous = check.continuous(pd.Series(joint))

    # Compute histograms of the data, bin if continuous
    if is_continuous:
        # Compute shared bin_edges if not given, and use np.histogram to form histograms
        if bin_edges is None:
            bin_edges = np.histogram_bin_edges(joint, bins="auto")

        hists = [np.histogram(series, bins=bin_edges)[0] for series in data]

        if normalize:
            with np.errstate(divide="ignore", invalid="ignore"):
                hists = [np.nan_to_num(hist / hist.sum()) for hist in hists]

    else:
        # For categorical data, form histogram using value counts and align
        space = joint.unique()

        dicts = [sr.value_counts(normalize=normalize) for sr in data]
        hists = [np.array([d.get(val, 0) for val in space]) for d in dicts]

    ps = [pd.Series(hist) for hist in hists]

    if ret_bins:
        return tuple(ps), bin_edges

    return tuple(ps)


def bootstrap_statistic(data: Union[Tuple[pd.Series], Tuple[pd.Series, pd.Series]],
                        statistic: Union[Callable[[pd.Series, pd.Series], float], Callable[[pd.Series], float]],
                        n_samples: int = 1000, sample_size=None) -> np.ndarray:
    """
    Compute the samples of a statistic estimate using the bootstrap method.

    Args:
        data: Data on which to compute the statistic.
        statistic: Function that computes the statistic.
        n_samples: Optional; Number of bootstrap samples to perform.

    Returns:
        The bootstrap samples.
    """
    if sample_size is None:
        sample_size = max((len(x) for x in data))

    def get_sample_idx(x):
        return np.random.randint(0, len(x), min(len(x), sample_size))

    statistic_samples = np.empty(n_samples)
    for i in range(n_samples):
        sample_idxs = [get_sample_idx(x) for x in data]
        statistic_samples[i] = statistic(*[x[idx] for x, idx in zip(data, sample_idxs)])
    return statistic_samples


def bootstrap_binned_statistic(data: Tuple[pd.Series, pd.Series], statistic: Callable[[pd.Series, pd.Series], float],
                               n_samples: int = 1000) -> np.ndarray:
    """
    Compute the samples of a binned statistic estimate using the bootstrap method.

    Args:
        data: Data for which to compute the statistic.
        statistic: Function that computes the statistic.
        n_samples: Optional; Number of bootstrap samples to perform.

    Returns:
        The bootstrap samples.
    """

    statistic_samples = np.empty(n_samples)

    with np.errstate(divide='ignore', invalid='ignore'):
        p_x = np.nan_to_num(data[0] / data[0].sum())
        p_y = np.nan_to_num(data[1] / data[1].sum())

    n_x = data[0].sum()
    n_y = data[1].sum()

    x_samples = np.random.multinomial(n_x, p_x, size=n_samples)
    y_samples = np.random.multinomial(n_y, p_y, size=n_samples)

    for i in range(n_samples):
        statistic_samples[i] = statistic(x_samples[i], y_samples[i])

    return statistic_samples



import copy

import numpy as np
import pandas as pd
import pytest

from synthesized_insight import ColumnCheck


@pytest.fixture(scope='module')
def df():
    df = pd.DataFrame({
        'string_col': np.random.choice(['A', 'B', 'C', 'D', 'E'], size=1000),
        'bool_col': np.random.choice([False, True], size=1000).astype('?'),
        'date_col': pd.to_datetime(18_000 + np.random.normal(500, 50, size=1000).astype(int), unit='D'),
        'int_col': np.random.randint(10, 100000, 1000),
        'float_col': np.random.normal(0.0, 1.0, size=1000),
        'int_bool_col': np.random.choice([0, 1], size=1000),
        'ordered_cat_col': pd.Categorical(np.random.choice(["b", "d", "c"], size=1000), categories=["b", "c", "d"], ordered=True)
    })

    return df


def verify_columns_types(df, categorical_cols, continuous_cols, affine_cols, ordinal_cols):
    check = ColumnCheck()

    pred_categorical_cols = set()
    for col in df.columns:
        if check.categorical(df[col]):
            pred_categorical_cols.add(col)
    assert pred_categorical_cols == categorical_cols

    pred_continuous_cols = set()
    for col in df.columns:
        if check.continuous(df[col]):
            pred_continuous_cols.add(col)
    assert pred_continuous_cols == continuous_cols

    pred_affine_cols = set()
    for col in df.columns:
        if check.affine(df[col]):
            pred_affine_cols.add(col)
    assert pred_affine_cols == affine_cols

    pred_ordinal_cols = set()
    for col in df.columns:
        if check.ordinal(df[col]):
            pred_ordinal_cols.add(col)
    assert pred_ordinal_cols == ordinal_cols


def test_column_check(df):
    categorical_cols = set(['string_col', 'bool_col', 'int_bool_col', 'ordered_cat_col'])
    continuous_cols = set(['int_col', 'float_col'])
    affine_cols = set(['date_col', 'int_col', 'float_col'])
    ordinal_cols = set(['ordered_cat_col', 'date_col', 'float_col', 'int_col', 'int_bool_col', 'bool_col'])

    verify_columns_types(df.copy(), categorical_cols, continuous_cols, affine_cols, ordinal_cols)

    # Adding some NaNs to the dataframe
    df_nan = copy.deepcopy(df)
    for col in df_nan.columns:
        df_nan.loc[df_nan.sample(frac=0.1).index, col] = np.nan

    verify_columns_types(df_nan, categorical_cols, continuous_cols, affine_cols, ordinal_cols)


def test_check_ordinal():
    check = ColumnCheck()

    sr = pd.Series([1, 2, 3, 4, 5])
    assert check.ordinal(sr) is True

    sr = pd.Series([3 for _ in range(100)] + [4 for _ in range(100)])
    assert check.ordinal(sr) is True

    sr = pd.Series([3 for _ in range(100)])
    assert check.ordinal(sr) is True

    sr = pd.Series(pd.Categorical(np.random.choice(["t2", "t1", "t0"], size=100), categories=["t0", "t1", "t2"], ordered=True))
    assert check.ordinal(sr) is True
